"""
Base classes for (file) I/O.

$Id: read.py 2747 2021-11-24 13:56:00Z pe $
"""

import re
from collections import defaultdict
from .base import EbasFileBase, isvalid
from nilutility.datatypes import DataObject
from ebas.domain.basic_domain_logic.unit_convert import UnitConvert, \
    NoConversion


class EbasFileRead(EbasFileBase):
    """
    Partial class for EbasFile (Base class for file objects).
    This part handles generic reading.
    """

    READ_EXCEPTION = NotImplementedError  # must be set in the concrete class

    def read(self, filename,
             ignore_rescode=False,
             ignore_revdate=False, ignore_parameter=False,
             ignore_sampleduration = False,
             ignore_identicalvalues=False,
             ignore_flagconsistency=False, fix_flagconsistency=False,
             ignore_valuecheck=False, ignore_templatecheck=False,
             skip_data=False, skip_unitconvert=False, skip_variables=None,
             fix_overlap_set_start=False, fix_bc211=False,
             **kwargs):
        """
        Reads a NASA Ames files to NasaAmes object.
        Parameters:
            filename          path and filename
            ignore_rescode    ignore rescde errors (downgraded to warning)
            ignore revdate    ignore revdate errors (downgrade to warning)
            ignore_parameters ignore errors related to paramters and units
                              this is needed when a file with non standard
                              vnames should be processed without importing it
                              into the ebas.domain.
            ignore_sampleduration
                              ignore errors related to the sample duration
                              metadata not being consistent with the data
            ignore_identicalvalues
                              ignore errors related to all values of one
                              variable being equal (e.g. all data missing for
                              one variable, all values 0.0 or all values any
                              other constant value).
            ignore_flagconsistency
                              ignore errors regarding flag consistency and
                              flag/value consitency
            fix_flagconsistency
                              ignore errors regarding flag consistency and
                              flag/value consitency; additionally fix problems
                              if possible
            ignore_valuecheck ignore value checks (boundaries and spikes)
            ignore_templatecheck ignore template checks
            skip_data         skip reading of data (speedup, if data are not
                              needed at application level). First and last line
                              are read. (Tjis is just passed to NasaAmes1001)
            skip_unitconvert  skips the automatic unit conversion on input
                              this is needed when a file with non standard units
                              should be processed without importing it into the
                              ebas.domain.
            skip_variables    list of variable numbers to be skipped (variabzle
                              numbers start with 1 for the first variable after
                              end_time)
            **kwargs          extra arguments for read_nasa_ames (e.g. encoding,
                              ignore_numformat, ignore_dx0)
            fix_overlap_set_start
                              instead of raising an error in case of overlaps,
                              set the previous end time to start time
            fix_bc211         flag boundary check violations with flag 211
        Returns:
            None
        Raises:
            IOError (from builtin, file open)
            EbasNasaAmesReadError
        """
        # initialize temporary attributes for reading
        self.internal.read_tmp = DataObject(
            var_skipped=[], var_num=[], vmiss=[], hea_error=False,
            revdate_read=False)

        self._read(
            filename, ignore_rescode=ignore_rescode,
            ignore_revdate=ignore_revdate, ignore_parameter=ignore_parameter,
            ignore_sampleduration=ignore_sampleduration,
            skip_data=skip_data,
            skip_variables=skip_variables, **kwargs)

        if len(self.sample_times) > 0:
            self.internal.read_tmp.timebounds = self.sample_times[0] + \
                                                self.sample_times[-1]
        else:
            self.internal.read_tmp.timebounds = None
        # skip data cleanup: now throw away first and last sample
        if skip_data:
            self.sample_times = []
            for var in self.variables:
                var.values_ = []
                var.flags = []

        self.set_default_metadata()  # must be before convert: regime needed!

        if not skip_unitconvert:
            # Unit convert must happen before check.
            # Otherwise parameter check would fail. 
            self.convert_variables_input()

        # check must be done in the end
        # e.g. unit convert on input changes unit, so that parameter/unit is
        # allowed
        self.check(ignore_rescode=ignore_rescode,
                   ignore_revdate=ignore_revdate,
                   ignore_parameter=ignore_parameter,
                   ignore_sampleduration=ignore_sampleduration,
                   ignore_identicalvalues=ignore_identicalvalues,
                   ignore_flagconsistency=ignore_flagconsistency,
                   fix_flagconsistency=fix_flagconsistency,
                   # ignore error about no data, when skipped:
                   ignore_nodata=skip_data,
                   ignore_valuecheck=ignore_valuecheck,
                   ignore_templatecheck=ignore_templatecheck,
                   fix_overlap_set_start=fix_overlap_set_start,
                   fix_bc211=fix_bc211)

        if self.errors > 0:
            self.logger.info("Exiting because of previous errors")
            raise self.__class__.READ_EXCEPTION(
                "{} Errors, {} Warnings".format(self.errors,
                                                self.warnings))

    def _read(self, filename,
              ignore_rescode=False,
              ignore_revdate=False, ignore_parameter=False, skip_data=False,
              skip_unitconvert=False, skip_variables=None, **kwargs):
        """
        Must be implemented in derived classes for the respective data format.
        Parameters:
            The implementation in derived classes can add parameters up to their
            neds. However, thoseparameters MUST be taken:

            filename          path and filename
            ignore_rescode    ignore rescde errors (downgraded to warning)
            ignore revdate    ignore revdate errors (downgrade to warning)
            ignore_parameters ignore errors related to paramters and units
                              this is needed when a file with non standard
                              vnames should be processed without importing it
                              into the ebas.domain.
            skip_data         skip reading of data (speedup, if data are not
                              needed at application level). First and last line
                              are read. (Tjis is just passed to NasaAmes1001)
            skip_unitconvert  skips the automatic unit conversion on input
                              this is needed when a file with non standard units
                              should be processed without importing it into the
                              ebas.domain.
            skip_variables    list of variable numbers to be skipped (variabzle
                              numbers start with 1 for the first variable after
                              end_time)
            **kwargs          extra arguments for read_nasa_ames (e.g. encoding,
                              ignore_numformat, ignore_dx0)
            Returns:
                None
            Raises:
                IOError (from builtin, file open)
                EbasNasaAmesReadError
        """
        raise NotImplementedError

    def convert_variables_input(self):
        """
        Convert variables on import (depending on component name, unit etc.
        E.g.: convert 'nitrogen_dioxide, nmol/mol' to
                      'nitrogen_dioxide, ug N/m3'
        Parameters:
            None
        Returns:
            None
        """
        self.internal.cvt_tmp = DataObject(
            global_dl_cand1=[], global_url_cand1=[], global_unc_cand1=[],
            global_qabias_cand1=defaultdict(list),
            global_qavariab_cand1=defaultdict(list),
            global_dl_cand2=[], global_url_cand2=[], global_unc_cand2=[],
            global_qabias_cand2=defaultdict(list),
            global_qavariab_cand2=defaultdict(list))
        for vnum in range(len(self.variables)):
            file_vnum, _, file_metalnum = self.msgref_vnum_lnum(vnum)
            regime = self.get_meta_for_var(vnum, 'regime')
            matrix = self.get_meta_for_var(vnum, 'matrix')
            comp_name = self.get_meta_for_var(vnum, 'comp_name')
            unit = self.get_meta_for_var(vnum, 'unit')
            std_pres = self.get_meta_for_var(vnum, 'vol_std_pressure')
            if not isinstance(std_pres, float):
                std_pres = None
            std_temp = self.get_meta_for_var(vnum, 'vol_std_temp')
            if not isinstance(std_temp, float):
                std_temp = None
            try:
                unit_convert = UnitConvert()
                (cvt, reset_stdcond) = unit_convert.import_conv_params(
                    regime, matrix, comp_name, unit, std_pres, std_temp)
            except NoConversion as expt:
                for msg in expt.msgs:
                    self.warning("variable {}: {}".format(
                        file_vnum, msg), lnum=file_metalnum)
            else:
                self._convert_variable_input(vnum, cvt)
                if reset_stdcond:
                    self.variables[vnum].metadata.vol_std_pressure = \
                        reset_stdcond[0]
                    self.variables[vnum].metadata.vol_std_temp = \
                        reset_stdcond[1]
        self._convert_global_metadata()

    def _check_convert_global_metadata(self, vnum, conv_obj):
        """
        Check if the global metadata detection limit (DL),
        upper range limit (URL) and uncertainty (UNC) can be unit converted
        with the same conversion as the variable vnum.
        This is the case, if the global parameter metadata are the same as the
        variable's, the global unit is the same as the variable's unit and the
        global DL, URL or UNC unit is also the same.
        In this case, the varaible's conversion object is a candidate for the
        conversion of global metadata DL, URL and UNC. The conversion of the
        global metadata is then performed by _convert_global_metadata().
        
        Parameters:
            vnum         variable number
            conv_obj     converison object
        Returns:
            None
        """
        regime = self.get_meta_for_var(vnum, 'regime')
        comp = self.get_meta_for_var(vnum, 'comp_name')
        matrix = self.get_meta_for_var(vnum, 'matrix')
        if self.metadata.regime == regime and \
           self.metadata.comp_name == comp and \
           self.metadata.matrix == matrix and \
           self.metadata.unit == conv_obj.from_unit:

            self.internal.cvt_tmp.global_unit = conv_obj.to_unit
            
            if 'uncertainty' in self.metadata and \
               isvalid(self.metadata.uncertainty) and \
               self.metadata.uncertainty[1] == conv_obj.from_unit:
                if "uncertainty" not in self.variables[vnum].metadata or \
                    self.metadata.uncertainty == \
                        self.variables[vnum].metadata.uncertainty:
                    # the global uncertainty will be used for this variable,
                    # so this is a first grade candidate (use the same rounding
                    # as for the variable:
                    self.internal.cvt_tmp.global_unc_cand1.append(conv_obj)
                else:
                    # the global uncertaionty is not used for this variable
                    # anyway it is a useable conversion candidate for global
                    # detection limit (with it's own rounding)
                    self.internal.cvt_tmp.global_unc_cand2.append(conv_obj)
            if 'detection_limit' in self.metadata and \
               isvalid(self.metadata.detection_limit) and \
               self.metadata.detection_limit[1] == conv_obj.from_unit:
                if "detection_limit" not in self.variables[vnum].metadata or\
                        self.metadata.detection_limit == \
                            self.variables[vnum].metadata.detection_limit:
                    # the global detection_limit will be used for this variable,
                    # so this is a first grade candidate (use the same rounding
                    # as for the variable:
                    self.internal.cvt_tmp.global_dl_cand1.append(conv_obj)
                else:
                    # the global detection_limit is not used for this variable
                    # anyway it is a useable conversion candidate for global
                    # detection limit (with it's own rounding)
                    self.internal.cvt_tmp.global_dl_cand2.append(conv_obj)
            if 'upper_range_limit' in self.metadata and \
               isvalid(self.metadata.upper_range_limit) and \
               self.metadata.upper_range_limit[1] == conv_obj.from_unit:
                if "upper_range_limit" not in self.variables[vnum].metadata or \
                        self.metadata.upper_range_limit == \
                            self.variables[vnum].metadata.upper_range_limit:
                    # the global UR limit will be used for this variable,
                    # so this is a first grade candidate (use the same rounding
                    # as for the variable:
                    self.internal.cvt_tmp.global_url_cand1.append(conv_obj)
                else:
                    # the global UR limit is not used for this variable
                    # anyway it is a useable conversion candidate for global
                    # detection limit (with it's own rounding)
                    self.internal.cvt_tmp.global_url_cand2.append(conv_obj)
            for qa_ in self.get_qa_for_var(vnum):
                if 'qa_bias' in qa_ and isvalid(qa_.qa_bias) and \
                        qa_.qa_bias[1] == conv_obj.from_unit:
                    varqa = self.get_qa_by_number(qa_.qa_number, vnum)
                    if 'qa_bias' not in varqa or varqa['qa_bias'] is None or \
                       varqa.qa_bias == qa_.qa_bias:
                        # the global qa bias will be used for this variable,
                        # so this is a first grade candidate (use the same
                        # rounding as for the variable:
                        self.internal.cvt_tmp.\
                            global_qabias_cand1[qa_.qa_number].append(conv_obj)
                    else:
                        # the global qa bias is not used for this variable
                        # anyway it is a useable conversion candidate for global
                        # detection limit (with it's own rounding)
                        self.internal.cvt_tmp.\
                            global_qabias_cand2[qa_.qa_number].append(conv_obj)
                if 'qa_variability' in qa_ and isvalid(qa_.qa_variability) and \
                       qa_.qa_variability[1] == conv_obj.from_unit:
                    varqa = self.get_qa_by_number(qa_.qa_number, vnum)
                    if 'qa_variability' not in varqa or \
                            varqa.qa_variability is None or \
                            varqa.qa_variability == qa_.qa_variability:
                        # the global qa variablility will be used for this
                        # variable, so this is a first grade candidate (use the
                        # same rounding as for the variable:
                        self.internal.cvt_tmp.\
                            global_qavariab_cand1[qa_.qa_number].append(
                                conv_obj)
                    else:
                        # the global qa variablility is not used for this
                        # variable anyway it is a useable conversion candidate
                        # for global detection limit (with it's own rounding)
                        self.internal.cvt_tmp.\
                            global_qavariab_cand2[qa_.qa_number].append(
                                conv_obj)

    def _convert_global_metadata(self):
        """
        Converts global metadata unit, uncertainty, detection limit and
        upper range limit.
        This is needed in those special cases:
            If the global metadata parameter and unit is set and the unit is
            converted for all variables, the global metadata remain with an
            invalid parameter/unit combination, which is a problem in
            check_param_unit() or the gloabal metadata.
            E.g. ozone file in nmol/mol, gloabl metadata ozone, nmol/mol, each
                 variable is ozone, nmol/mol.
                 After conversion, all valiables are ozone, ug/m3, but the
                 global parameter/unit check would fail.
            Once the global unit is converted accordingly, the check of global
            detection limit, upper range limit and uncertainty might fail, e.g.:
            ERROR   : Unit 'X' used for 'Detection limit' is inconsistent
                      with 'Unit' ('Y')

        to be documented in the code:
         - only when the global metadata is not needed anymore (i.e.
        all applicable variables have been converted and thus have got their
        own, variable specific metadata element.
        If the global metadata still applies to any variable, the variable has
        obviously not been converted and the global metadata is needed as was.
        This might lead to resulting errors in consistency within the global
        metadata...
        Parameters:
            conv_obj     conversion object
        Returns:
            None
        """
        if 'global_unit' in self.internal.cvt_tmp:
            self.logger.info("file global unit '{}' changed to '{}'".format(
                self.metadata.unit, self.internal.cvt_tmp.global_unit))
            self.metadata.unit = self.internal.cvt_tmp.global_unit

        if self.internal.cvt_tmp.global_unc_cand1 or \
           self.internal.cvt_tmp.global_unc_cand2:
            if isvalid(self.metadata.uncertainty):
                used_var = [vnum + 1
                            for vnum in range(len(self.variables))
                            if 'uncertainty' not in \
                                   self.variables[vnum].metadata and \
                               self.variables[vnum].metadata.unit == \
                                   self.metadata.uncertainty[1]]
                if used_var:
                    self.warning(
                        "File global 'Uncertaionty' cannot be unit converted "
                        "because it is still used in variable{} {}".format(
                            "s" if len(used_var) > 1 else "",
                            ', '.join(used_var)))
                else:
                    reset = False
                    if self.internal.cvt_tmp.global_unc_cand1:
                        if len(set([x.rounding
                                    for x in \
                                    self.internal.cvt_tmp.global_unc_cand1])) \
                            == 1:
                            # all candidates use the same rounding -- use the
                            # same
                            conv_obj = self.internal.cvt_tmp.global_unc_cand1[0]
                        else:
                            # use the first candidate, but reset the rounding
                            conv_obj = self.internal.cvt_tmp.global_unc_cand1[0]
                            reset = conv_obj.rounding
                    else:
                        # use the first candidate from global_unc_cand2 and
                        # reset rounding
                        conv_obj = self.internal.cvt_tmp.global_unc_cand2[0]
                        reset = conv_obj.rounding

                    origval = self.metadata.uncertainty[0]
                    value_list = [origval]
                    conv_obj.convert_data(value_list)
                    newval = value_list[0]
                    self.metadata.uncertainty = (newval, conv_obj.to_unit)
                    cnvstr = conv_obj.conversion_string(from_val=origval,
                                                        to_val=newval)
                    self.warning(
                        "Converting file global 'Measurement uncertainty' for "
                        "component {} {}.".format(self.metadata.comp_name,
                                                  cnvstr))
                    if reset != False:
                        conv_obj.rounding = reset  # reset the rounding

        if self.internal.cvt_tmp.global_dl_cand1 or \
           self.internal.cvt_tmp.global_dl_cand2:
            if isvalid(self.metadata.detection_limit):
                used_var = [vnum + 1
                            for vnum in range(len(self.variables))
                            if 'detection_limit' not in \
                                   self.variables[vnum].metadata and \
                               self.variables[vnum].metadata.unit == \
                                   self.metadata.detection_limit[1]]
                if used_var:
                    self.warning(
                        "File global 'Detection limit' cannot be unit "
                        "converted because it is still used in variable{} {}"
                        .format(
                            "s" if len(used_var) > 1 else "",
                            ', '.join(used_var)))
                else:
                    reset = False
                    if self.internal.cvt_tmp.global_dl_cand1:
                        if len(set([x.rounding
                                    for x in \
                                    self.internal.cvt_tmp.global_dl_cand1])) \
                            == 1:
                            # all candidates use the same rounding -- use the
                            # same
                            conv_obj = self.internal.cvt_tmp.global_dl_cand1[0]
                        else:
                            # use the first candidate, but reset the rounding
                            conv_obj = self.internal.cvt_tmp.global_dl_cand1[0]
                            reset = conv_obj.rounding
                    else:
                        # use the first candidate from global_dl_cand2 and reset
                        # rounding
                        conv_obj = self.internal.cvt_tmp.global_dl_cand2[0]
                        reset = conv_obj.rounding

                    origval = self.metadata.detection_limit[0]
                    value_list = [origval]
                    conv_obj.convert_data(value_list)
                    newval = value_list[0]
                    self.metadata.detection_limit = (newval, conv_obj.to_unit)
                    cnvstr = conv_obj.conversion_string(from_val=origval,
                                                        to_val=newval)
                    self.warning(
                        "Converting file global 'Detection limit' for "
                        "component {} {}.".format(self.metadata.comp_name,
                                                  cnvstr))
                    if reset != False:
                        conv_obj.rounding = reset  # reset the rounding

        if self.internal.cvt_tmp.global_url_cand1 or \
           self.internal.cvt_tmp.global_url_cand2:
            if isvalid(self.metadata.upper_range_limit):
                used_var = [vnum + 1
                            for vnum in range(len(self.variables))
                            if 'upper_range_limit' not in \
                                   self.variables[vnum].metadata and \
                               self.variables[vnum].metadata.unit == \
                                   self.metadata.upper_range_limit[1]]
                if used_var:
                    self.warning(
                        "File global 'Upper range limit' cannot be unit "
                        "converted because it is still used in variable{} {}"
                        .format(
                            "s" if len(used_var) > 1 else "",
                            ', '.join(used_var)))
                else:
                    reset = False
                    if self.internal.cvt_tmp.global_url_cand1:
                        if len(set([x.rounding
                                    for x in \
                                    self.internal.cvt_tmp.global_url_cand1])) \
                            == 1:
                            # all candidates use the same rounding -- use the
                            # same
                            conv_obj = self.internal.cvt_tmp.global_url_cand1[0]
                        else:
                            # use the first candidate, but reset the rounding
                            conv_obj = self.internal.cvt_tmp.global_url_cand1[0]
                            reset = conv_obj.rounding
                    else:
                        # use the first candidate from global_url_cand2 and
                        # reset rounding
                        conv_obj = self.internal.cvt_tmp.global_url_cand2[0]
                        reset = conv_obj.rounding

                    origval = self.metadata.upper_range_limit[0]
                    value_list = [origval]
                    conv_obj.convert_data(value_list)
                    newval = value_list[0]
                    self.metadata.upper_range_limit = (newval, conv_obj.to_unit)
                    cnvstr = conv_obj.conversion_string(from_val=origval,
                                                        to_val=newval)
                    self.warning(
                        "Converting file global 'Upper range limit' for "
                        "component {} {}.".format(self.metadata.comp_name,
                                                  cnvstr))
                    if reset != False:
                        conv_obj.rounding = reset  # reset the rounding

        for qanum in self.get_qa_numbers():
            qa_ = self.get_qa_by_number(qanum, None)

            if self.internal.cvt_tmp.global_qabias_cand1[qanum] or \
               self.internal.cvt_tmp.global_qabias_cand2[qanum]:
                if 'qa_bias' in qa_ and isvalid(qa_.qa_bias):
                    used_var = []
                    for vnum in range(len(self.variables)):
                        varqa = self.get_qa_by_number(qanum, vnum)
                        if ('qa_bias' not in varqa or varqa.qa_bias is None) \
                                and self.variables[vnum].metadata.unit == \
                                    qa_.qa_bias[1]:
                            used_var.append(vnum + 1)
                    if used_var:
                        self.warning(
                            "File global 'QA{} bias' cannot be unit converted "
                            "because it is still used in variable{} {}"
                            .format(
                                str(qanum) if qanum != -1 else '',
                                "s" if len(used_var) > 1 else "",
                                ', '.join(used_var)))
                    else:
                        reset = False
                        if self.internal.cvt_tmp.global_qabias_cand1[qanum]:
                            if len(set(
                                    [x.rounding
                                     for x in \
                                     self.internal.cvt_tmp.\
                                     global_qabias_cand1[qanum]])) == 1:
                                # all candidates use the same rounding -- use
                                # the same
                                conv_obj = self.internal.cvt_tmp.\
                                    global_qabias_cand1[qanum][0]
                            else:
                                # use the first candidate, but reset the
                                # rounding
                                conv_obj = self.internal.cvt_tmp.\
                                    global_qabias_cand1[qanum][0]
                                reset = conv_obj.rounding
                        else:
                            # use the first candidate from global_qabias_cand2
                            # and reset rounding
                            conv_obj = self.internal.cvt_tmp.\
                                global_qabias_cand2[qanum][0]
                            reset = conv_obj.rounding

                        origval = qa_.qa_bias[0]
                        value_list = [origval]
                        conv_obj.convert_data(value_list)
                        newval = value_list[0]
                        qa_.qa_bias = (newval, conv_obj.to_unit)
                        cnvstr = conv_obj.conversion_string(from_val=origval,
                                                            to_val=newval)
                        self.warning(
                            "Converting file global 'QA{} bias' for "
                            "component {} {}.".format(
                                str(qanum) if qanum != -1 else '',
                                self.metadata.comp_name,
                                cnvstr))
                        if reset != False:
                            conv_obj.rounding = reset  # reset the rounding

            if self.internal.cvt_tmp.global_qavariab_cand1[qanum] or \
               self.internal.cvt_tmp.global_qavariab_cand2[qanum]:
                if 'qa_variability' in qa_ and isvalid(qa_.qa_variability):
                    used_var = []
                    for vnum in range(len(self.variables)):
                        varqa = self.get_qa_by_number(qanum, vnum)
                        if ('qa_variability' not in varqa or \
                                    varqa.qa_variability is None) and \
                                self.variables[vnum].metadata.unit == \
                                    qa_.qa_variability[1]:
                            used_var.append(vnum + 1)
                    if used_var:
                        self.warning(
                            "File global 'QA{} variability' cannot be unit "
                            "converted because it is still used in "
                            "variable{} {}".format(
                                str(qanum) if qanum != -1 else '',
                                "s" if len(used_var) > 1 else "",
                                ', '.join(used_var)))
                    else:
                        reset = False
                        if self.internal.cvt_tmp.global_qavariab_cand1[qanum]:
                            if len(set(
                                    [x.rounding
                                     for x in \
                                     self.internal.cvt_tmp.\
                                        global_qavariab_cand1[qanum]])) == 1:
                                # all candidates use the same rounding -- use
                                # the same
                                conv_obj = self.internal.cvt_tmp.\
                                    global_qavariab_cand1[qanum][0]
                            else:
                                # use the first candidate, but reset the
                                # rounding
                                conv_obj = self.internal.cvt_tmp.\
                                    global_qavariab_cand1[qanum][0]
                                reset = conv_obj.rounding
                        else:
                            # use the first candidate from global_qavariab_cand2
                            # and reset rounding
                            conv_obj = self.internal.cvt_tmp.\
                                global_qavariab_cand2[qanum][0]
                            reset = conv_obj.rounding

                        origval = qa_.qa_variability[0]
                        value_list = [origval]
                        conv_obj.convert_data(value_list)
                        newval = value_list[0]
                        qa_.qa_variability = (newval, conv_obj.to_unit)
                        cnvstr = conv_obj.conversion_string(from_val=origval,
                                                            to_val=newval)
                        self.warning(
                            "Converting file global 'QA{} variability' for "
                            "component {} {}.".format(
                                str(qanum) if qanum != -1 else '',
                                self.metadata.comp_name,
                                cnvstr))
                        if reset != False:
                            conv_obj.rounding = reset  # reset the rounding

    def _convert_variable_input(self, vnum, conv_obj):
        """
        Convert one variables on import from one unit to another.
        E.g.: convert 'nitrogen_dioxide, nmol/mol' to
                      'nitrogen_dioxide, ug N/m3'
        Parameters:
            vnum         variable number
            conv_obj     converison object
        Returns:
            None
        """
        file_vnum, _, file_metalnum = self.msgref_vnum_lnum(vnum)
        comp = self.get_meta_for_var(vnum, 'comp_name')
        self._check_convert_global_metadata(vnum, conv_obj)

        # first, convert the data, to get the rounding right according to data:
        if len(self.sample_times) < 100 and \
                'read_tmp' in self.internal and self.internal.read_tmp and \
                'vmiss' in self.internal.read_tmp and \
                self.internal.read_tmp.vmiss and \
                re.match(r'^9+(\.9*)?$', self.internal.read_tmp.vmiss[vnum]):
            # special case for input conversion if not enough values:
            # use missing value in addition to esimate precision
            # (only 99.99 notation missing values can be used,
            # 0x are not needed and 9E9 format does not help for rounding....
            self.variables[vnum].values_.append(
                    float(self.internal.read_tmp.vmiss[vnum]))
            conv_obj.convert_data(self.variables[vnum].values_)
            del self.variables[vnum].values_[-1]
        else:
            conv_obj.convert_data(self.variables[vnum].values_)

        self.variables[vnum].metadata.unit = conv_obj.to_unit  # change unit

        cnvstr = conv_obj.conversion_string()
        self.warning("variable {}: Component '{}', converting unit {}."
                      .format(file_vnum, comp, cnvstr), lnum=file_metalnum)
        conv_remark = "Data converted on import into EBAS {}.".format(cnvstr)

        # change uncertainty, detection limit and upper range limit, qa_qa_bias
        # and qa_variability
        if 'uncertainty' in self.variables[vnum].metadata and \
                isvalid(self.variables[vnum].metadata.uncertainty) and \
                self.variables[vnum].metadata.uncertainty[1] == \
                    conv_obj.from_unit:
            uncert = self.variables[vnum].metadata.uncertainty
            if conv_obj.rounding is None:
                # rounding needs to be reset after conversion
                reset = True
            else:
                reset = False
            origval = uncert[0]
            value_list = [origval]
            conv_obj.convert_data(value_list)
            newval = value_list[0]
            self.variables[vnum].metadata.uncertainty = (newval,
                                                         conv_obj.to_unit)
            cnvstr = conv_obj.conversion_string(from_val=origval, to_val=newval)
            self.warning(
                "variable {}: converting 'Measurement uncertainty' for "
                "component {} {}.".format(file_vnum, comp, cnvstr),
                lnum=file_metalnum)
            conv_remark += " Measurement uncertainty converted {}.".format(
                cnvstr)
            if reset:
                conv_obj.rounding = None  # reset the rounding
        if 'detection_limit' in self.variables[vnum].metadata and \
                isvalid(self.variables[vnum].metadata.detection_limit) and \
                self.variables[vnum].metadata.detection_limit[1] == \
                    conv_obj.from_unit:
            detect = self.variables[vnum].metadata.detection_limit
            if conv_obj.rounding is None:
                # rounding needs to be reset after conversion
                reset = True
            else:
                reset = False
            origval = detect[0]
            value_list = [origval]
            conv_obj.convert_data(value_list)
            newval = value_list[0]
            self.variables[vnum].metadata.detection_limit = (newval,
                                                             conv_obj.to_unit)
            cnvstr = conv_obj.conversion_string(from_val=origval, to_val=newval)
            self.warning(
                "variable {}: converting 'Detection limit' for component {} {}."
                .format(file_vnum, comp, cnvstr), lnum=file_metalnum)
            conv_remark += " Detection limit converted {}.".format(cnvstr)
            if reset:
                conv_obj.rounding = None  # reset the rounding
        if 'upper_range_limit' in self.variables[vnum].metadata and \
                isvalid(self.variables[vnum].metadata.upper_range_limit) and \
                self.variables[vnum].metadata.upper_range_limit[1] == \
                    conv_obj.from_unit:
            upper = self.variables[vnum].metadata.upper_range_limit
            if conv_obj.rounding is None:
                # rounding needs to be reset after conversion
                reset = True
            else:
                reset = False
            origval = upper[0]
            value_list = [origval]
            conv_obj.convert_data(value_list)
            newval = value_list[0]
            self.variables[vnum].metadata.upper_range_limit = (newval,
                                                               conv_obj.to_unit)
            cnvstr = conv_obj.conversion_string(from_val=origval, to_val=newval)
            self.warning(
                "variable {}: converting 'Upper range limit' for component "
                "{} {}." .format(file_vnum, comp, cnvstr),
                lnum=file_metalnum)
            conv_remark += " Upper range limit converted {}.".format(cnvstr)
            if reset:
                conv_obj.rounding = None  # reset the rounding
        for qanum in self.get_qa_numbers():
            qa_ = self.get_qa_by_number(qanum, vnum)
            if 'qa_bias' in qa_ and qa_.qa_bias[1] == conv_obj.from_unit:
                if conv_obj.rounding is None:
                    # rounding needs to be reset after conversion
                    reset = True
                else:
                    reset = False
                origval = qa_.qa_bias[0]
                value_list = [origval]
                conv_obj.convert_data(value_list)
                newval = value_list[0]
                qa_.qa_bias = (newval, conv_obj.to_unit)
                cnvstr = conv_obj.conversion_string(from_val=origval,
                                                    to_val=newval)
                self.warning(
                    "variable {}: converting 'QA{} bias' for component "
                    "{} {}." .format(
                        file_vnum,
                        str(qa_.qa_number) if qa_.qa_number != -1 else '',
                        comp, cnvstr),
                    lnum=file_metalnum)
                conv_remark += " QA{} bias converted {}.".format(
                    str(qa_.qa_number) if qa_.qa_number != -1 else '',
                    cnvstr)
                if reset:
                    conv_obj.rounding = None  # reset the rounding
            if 'qa_variability' in qa_ and \
                    qa_.qa_variability[1] == conv_obj.from_unit:
                if conv_obj.rounding is None:
                    # rounding needs to be reset after conversion
                    reset = True
                else:
                    reset = False
                origval = qa_.qa_variability[0]
                value_list = [origval]
                conv_obj.convert_data(value_list)
                newval = value_list[0]
                qa_.qa_variability = (newval, conv_obj.to_unit)
                cnvstr = conv_obj.conversion_string(from_val=origval,
                                                    to_val=newval)
                self.warning(
                    "variable {}: converting 'QA{} variability' for component "
                    "{} {}." .format(
                        file_vnum,
                        str(qa_.qa_number) if qa_.qa_number != -1 else '',
                        comp, cnvstr),
                    lnum=file_metalnum)
                conv_remark += " QA{} variability converted {}.".format(
                    str(qa_.qa_number) if qa_.qa_number != -1 else '',
                    cnvstr)
                if reset:
                    conv_obj.rounding = None  # reset the rounding

        # add comment
        comment = self.get_meta_for_var(vnum, 'comment')
        if comment:
            comment += " -- " + conv_remark
        else:
            comment = conv_remark
        self.variables[vnum].metadata.comment = comment



"""
$Id: sc.py 2088 2018-06-29 13:46:38Z pe $

EBAS Masterdata Class for statistics codes

This module implements the class EbasMasterSC.
"""

from .offline_masterdata import SCOfflineMasterData
from .base import EbasMasterBase

class EbasMasterSC(EbasMasterBase, SCOfflineMasterData):
    """
    Domain Class for statistics codes masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling statistics codes and checking them against master data.
    Statistics codes master data are retrieved from database or from offline
    storage when no database access is possible.
    """

    # read offline masterdata for SC (statistcs codes)
    # Those are fallback values, will be read from database as soon as possible.
    SCOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read sc masterdata if dbh is provided.
        """
        SCOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)



"""
$Id: fp.py 2088 2018-06-29 13:46:38Z pe $

EBAS Masterdata Class for filter prefiring masterdata

This module implements the class EbasMasterFP.

History:
V.1.1.0  2016-10-17  pe  initial version

"""

from .offline_masterdata import FPOfflineMasterData
from .base import EbasMasterBase

class EbasMasterFP(EbasMasterBase, FPOfflineMasterData):
    """
    Domain Class for filter prefiring masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling stations and checking them against master data.
    Master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for PR (projects)
    # Those are fallback values, will be read from database as soon as possible.
    FPOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read fp masterdata if dbh is provided.
        """
        FPOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)



"""
$Id: it.py 2088 2018-06-29 13:46:38Z pe $

EBAS Masterdata Class for inlet type masterdata

This module implements the class EbasMasterIT.

"""

from .offline_masterdata import ITOfflineMasterData
from .base import EbasMasterBase

class EbasMasterIT(EbasMasterBase, ITOfflineMasterData):
    """
    Domain Class for inlet type masterdata
    Objects of this class do not represent entities, this class provides class
    methods for handling components and checking them against master data.
    Components master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for IT (inlet type)
    # Those are fallback values, will be read from database as soon as possible.
    ITOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read IT masterdata if dbh is provided.
        """
        ITOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)




"""
ebas/nasa_ames/read/legacy_vnames.py
$Id: legacy_vnames.py 1728 2017-07-11 09:00:19Z pe $

Translator for legacy variable names

History:
V.1.0.0  2013-06-22  pe  initial version

"""

import re
from .parse_base import NasaAmesPartialReadParserBase

class NasaAmesPartialReadLegacyVnames(# pylint: disable=R0901, W0223,
        NasaAmesPartialReadParserBase):
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Translator for legacy variable names.
    This is a base class for NasaAmesPartialReadParserVariables
    """

    def translate_legacy_vnames(self):
        """
        Translates legacy VNAME lines to current (EBAS_1.1) standard.
        Parameters:
            None
        Returns:
            None
        Raises:
        Variable number in the object differs from variable number in file
        (because the object only stores measurements as variables, having
        flags as an attribute)
        """
        for vnum in range(self.nasa1001.data.NV):
            if vnum in self.internal.read_tmp.var_skipped:
                # already skipped?
                continue
            vname = self.nasa1001.data.VNAME[vnum]
            for tran in self.__class__.Translations:
                if re.search(tran[0], vname):
                    if tran[1] is None:
                        self.warning(
                            "VNAME[{}]: Legacy VNAME '{}' skipped".format(
                                vnum, vname), lnum=13+vnum)
                        self.internal.read_tmp.var_skipped.append(vnum)
                        break
                    vname = re.sub(tran[0], tran[1], vname)
            if vnum in self.internal.read_tmp.var_skipped:
                # now skipped?
                continue
            if self.nasa1001.data.VNAME[vnum] != vname:
                self.warning(
                    "VNAME[{}]: Legacy VNAME '{}' translated to '{}'".format(
                        vnum, self.nasa1001.data.VNAME[vnum], vname),
                    lnum=13+vnum)

            # EBAS_1 legacy: VNAME 'value'
            if re.match('^ *value *,', vname):
                if self.metadata.datadef == 'EBAS_1' and \
                   self.nasa1001.data.NV == 3:
                    if self.metadata.comp_name is None:
                        self.error("VNAME[{}]: ".format(vnum) +\
                            "generic VNAME 'value' can't be used when Metadata "
                            "'Component' is missing in NCOM", lnum=13+vnum)
                    else:
                        vname = re.sub('^ *value *', self.metadata.comp_name,
                                       vname)
                elif self.nasa1001.data.NV == 3:
                    if self.metadata.comp_name is None:
                        self.error(
                            "VNAME[{}]: legacy EBAS_1 VNAME 'value' can't be "
                            "used when Metadata 'Component' is missing in NCOM".
                            format(vnum), lnum=13+vnum)
                    else:
                        vname = re.sub('^ *value *', self.metadata.comp_name,
                                       vname, 13 + vnum)

                        self.warning(
                            "VNAME[{}]: legacy EBAS_1 VNAME 'value', "
                            "translated to '{}'".format(
                                vnum, self.metadata.comp_name),
                            lnum=13+vnum)
                else:
                    self.error("VNAME[{}]: ".format(vnum) +\
                        "VNAME 'value' is illegal for multicolumn files; "
                        "variable skipped", lnum=13+vnum)
                    self.internal.read_tmp.var_skipped.append(vnum)
                    continue # stop parsing this line

            if vname == 'numflag, no unit, max 3 flags (of 3 digits each) ' +\
                        'coded into decimals':
                if self.metadata.datadef == 'EBAS_1':
                    vname = 'numflag, no unit'
                # else: gnererate a "unconventional VNAME for flag column" error
                # message in _parse_variable_numflag()
            if re.search('nominal/measured *=', vname):
                vname = re.sub('nominal/measured *=', 'Nominal/measured=',
                               vname)
                self.warning(
                    "VNAME[{}]: Legacy characteristic 'nominal/measured' "
                    "translated to 'Nominal/measured'".format(vnum),
                    lnum=13+vnum)

            self.nasa1001.data.VNAME[vnum] = vname

    Translations = (
        (r'^start_time of measurement, *year *$', None),
        (r'^end_time of measurement, *year *$', None),
        (r'^number of wavelengths\s*$', None),
        (r'^number of wavelengths,\s*no unit', None),

        (r'^number of size bins *$', None),
        (r'^number of size bins,\s*no unit', None),
        (r'^absorption_Aangstroem_coefficient, no unit$', None),

        (r'^([^, ]*)_statistics', '\\1'),

        (r'^internal instrument pressure, *hPa',
         'pressure, hPa, Location=instrument internal'),
        (r'^Instrument internal Pressure, *hPa',
         'pressure, hPa, Location=instrument internal'),

        (r'^internal instrument temperature, *K',
         'temperature, K, Location=instrument internal'),
        (r'^Instrument internal Temperature, *K',
         'temperature, K, Location=instrument internal'),
        (r'^sample temperature measured at inlet, *K',
         'temperature, K, Location=inlet'),
        (r'^sample temperature measured at outlet, *K',
         'temperature, K, Location=outlet'),

        (r'^internal relative humidity, *%',
         'relative_humidity, %, Location=instrument internal'),
        (r'^Instrument internal relative humidity, *%',
         'relative_humidity, %, Location=instrument internal'),
        (r'^relative humidity measured at inlet, *%',
         'relative_humidity, %, Location=inlet'),
        (r'^relative humidity measured at outlet, *%',
         'relative_humidity, %, Location=outlet'),

        (r'^lamp supply voltage, V',
         'electric_tension, V, Matrix=instrument, '
         'Location=lamp lamp supply'),
        (r'^lamp supply current, A',
         'electric_current, A, Matrix=instrument, '
         'Location=lamp lamp supply'),
        (r'^status flags, hex notation',
         'status, no unit, Matrix=instrument, '
         'Status type=overall instrument status'),

        (r'^recalibrated, truncation corrected ((back)?scattering) ' +\
             r'coefficient at (\d+) nm, 1/Mm',
         'aerosol_light_\\1_coefficient, 1/Mm, Wavelength=\\3nm'),
        (r'^((back)?scattering) coefficient (\d+\.\d*) percentile at (\d+) nm, '
         '1/Mm',
         'aerosol_light_\\1_coefficient, 1/Mm, Wavelength=\\4nm, '
         'Statistics=percentile:\\3'),

        (r'^(aerosol )?absorption coefficient at (\d+) ?nm, 1/Mm',
         'aerosol_absorption_coefficient, 1/Mm, Wavelength=\\2nm'),
        (r'^(aerosol )?absorption coefficient at (\d+) ?nm, (\d+\.\d*)' +\
              r'percentile, 1/Mm',
         'aerosol_absorption_coefficient, 1/Mm, Wavelength=\\2nm, '
         'Statistics=percentile:\\3'),
        (r'^(aerosol )?absorption coefficient (\d+\.\d*) percentile at ' +\
             r'(\d+) ?nm, 1/Mm',
         'aerosol_absorption_coefficient, 1/Mm, Wavelength=\\3nm, '
         'Statistics=percentile:\\2'),

        (r'^absorption_coefficient, 1/Mm',
         'aerosol_absorption_coefficient, 1/Mm'),

        (r'^aerosol number concentration dN/dlogD at D= *(\d+\.\d*) *nm, ' +\
             r'1/cm3',
         'particle_number_size_distribution, 1/cm3, D=\\1nm'),
        (r'^sample flow rate, standard conditions', 'flow_rate'),
        (r'^aerosol flow rate, ', 'flow_rate, '),

        (r', *Arithmetic mean', ', Statistics=arithmetic mean'),
        (r', *Median', ', Statistics=median'),
        (r', *Stddev', ', Statistics=stddev'),

        ('aerosol_light_backscattering_coefficient_statistics',
         'aerosol_light_backscattering_coefficient'),

        ('global_radiation',
         'downward_solar_radiation_flux_density'),

        )



"""
NOAA NMHC Flask data file format

$Id: __init__.py 2725 2021-10-29 07:04:53Z pe $
"""

from .base import EanetError
from .read_filter import EanetFilterRead
from .read_precip import EanetPrecipRead

class EanetFilter(EanetFilterRead):
    """
    The final class EanetFilter inherits from read and possible more in the
    future?
    """
    pass

class EanetPrecip(EanetPrecipRead):
    """
    The final class EanetPrecip inherits from read and possible more in the
    future?
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Generated Tue Feb 16 21:30:52 2016 by generateDS.py version 2.19a.
#
# Command line options:
#   ('--external-encoding', 'utf-8')
#   ('--no-process-includes', '')
#   ('-o', 'ebas_diffexport_generated.py')
#
# Command line arguments:
#   schema/ebas_diffexport.xsd
#
# Command line:
#   /Library/Frameworks/Python.framework/Versions/2.7/bin/generateDS.py --external-encoding="utf-8" --no-process-includes -o "ebas_diffexport_generated.py" schema/ebas_diffexport.xsd
#
# Current working directory (os.getcwd()):
#   generated
#

import sys
import re as re_
import base64
import datetime as datetime_
import warnings as warnings_
from six import string_types
from lxml import etree as etree_


Validate_simpletypes_ = True


def parsexml_(infile, parser=None, **kwargs):
    if parser is None:
        # Use the lxml ElementTree compatible parser so that, e.g.,
        #   we ignore comments.
        parser = etree_.ETCompatXMLParser()
    doc = etree_.parse(infile, parser=parser, **kwargs)
    return doc

#
# User methods
#
# Calls to the methods in these classes are generated by generateDS.py.
# You can replace these methods by re-implementing the following class
#   in a module named generatedssuper.py.

try:
    from generatedssuper import GeneratedsSuper
except ImportError as exp:

    class GeneratedsSuper(object):
        tzoff_pattern = re_.compile(r'(\+|-)((0\d|1[0-3]):[0-5]\d|14:00)$')
        class _FixedOffsetTZ(datetime_.tzinfo):
            def __init__(self, offset, name):
                self.__offset = datetime_.timedelta(minutes=offset)
                self.__name = name
            def utcoffset(self, dt):
                return self.__offset
            def tzname(self, dt):
                return self.__name
            def dst(self, dt):
                return None
        def gds_format_string(self, input_data, input_name=''):
            return input_data
        def gds_validate_string(self, input_data, node=None, input_name=''):
            if not input_data:
                return ''
            else:
                return input_data
        def gds_format_base64(self, input_data, input_name=''):
            return base64.b64encode(input_data)
        def gds_validate_base64(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_integer(self, input_data, input_name=''):
            return '%d' % input_data
        def gds_validate_integer(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_integer_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_integer_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    int(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of integers')
            return values
        def gds_format_float(self, input_data, input_name=''):
            return ('%.15f' % input_data).rstrip('0')
        def gds_validate_float(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_float_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_float_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    float(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of floats')
            return values
        def gds_format_double(self, input_data, input_name=''):
            return '%e' % input_data
        def gds_validate_double(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_double_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_double_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                try:
                    float(value)
                except (TypeError, ValueError):
                    raise_parse_error(node, 'Requires sequence of doubles')
            return values
        def gds_format_boolean(self, input_data, input_name=''):
            return ('%s' % input_data).lower()
        def gds_validate_boolean(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_boolean_list(self, input_data, input_name=''):
            return '%s' % ' '.join(input_data)
        def gds_validate_boolean_list(
                self, input_data, node=None, input_name=''):
            values = input_data.split()
            for value in values:
                if value not in ('true', '1', 'false', '0', ):
                    raise_parse_error(
                        node,
                        'Requires sequence of booleans '
                        '("true", "1", "false", "0")')
            return values
        def gds_validate_datetime(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_datetime(self, input_data, input_name=''):
            if input_data.microsecond == 0:
                _svalue = '%04d-%02d-%02dT%02d:%02d:%02d' % (
                    input_data.year,
                    input_data.month,
                    input_data.day,
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                )
            else:
                _svalue = '%04d-%02d-%02dT%02d:%02d:%02d.%s' % (
                    input_data.year,
                    input_data.month,
                    input_data.day,
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                    ('%f' % (float(input_data.microsecond) / 1000000))[2:],
                )
            if input_data.tzinfo is not None:
                tzoff = input_data.tzinfo.utcoffset(input_data)
                if tzoff is not None:
                    total_seconds = tzoff.seconds + (86400 * tzoff.days)
                    if total_seconds == 0:
                        _svalue += 'Z'
                    else:
                        if total_seconds < 0:
                            _svalue += '-'
                            total_seconds *= -1
                        else:
                            _svalue += '+'
                        hours = total_seconds // 3600
                        minutes = (total_seconds - (hours * 3600)) // 60
                        _svalue += '{0:02d}:{1:02d}'.format(hours, minutes)
            return _svalue
        @classmethod
        def gds_parse_datetime(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            time_parts = input_data.split('.')
            if len(time_parts) > 1:
                micro_seconds = int(float('0.' + time_parts[1]) * 1000000)
                input_data = '%s.%s' % (time_parts[0], micro_seconds, )
                dt = datetime_.datetime.strptime(
                    input_data, '%Y-%m-%dT%H:%M:%S.%f')
            else:
                dt = datetime_.datetime.strptime(
                    input_data, '%Y-%m-%dT%H:%M:%S')
            dt = dt.replace(tzinfo=tz)
            return dt
        def gds_validate_date(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_date(self, input_data, input_name=''):
            _svalue = '%04d-%02d-%02d' % (
                input_data.year,
                input_data.month,
                input_data.day,
            )
            try:
                if input_data.tzinfo is not None:
                    tzoff = input_data.tzinfo.utcoffset(input_data)
                    if tzoff is not None:
                        total_seconds = tzoff.seconds + (86400 * tzoff.days)
                        if total_seconds == 0:
                            _svalue += 'Z'
                        else:
                            if total_seconds < 0:
                                _svalue += '-'
                                total_seconds *= -1
                            else:
                                _svalue += '+'
                            hours = total_seconds // 3600
                            minutes = (total_seconds - (hours * 3600)) // 60
                            _svalue += '{0:02d}:{1:02d}'.format(
                                hours, minutes)
            except AttributeError:
                pass
            return _svalue
        @classmethod
        def gds_parse_date(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            dt = datetime_.datetime.strptime(input_data, '%Y-%m-%d')
            dt = dt.replace(tzinfo=tz)
            return dt.date()
        def gds_validate_time(self, input_data, node=None, input_name=''):
            return input_data
        def gds_format_time(self, input_data, input_name=''):
            if input_data.microsecond == 0:
                _svalue = '%02d:%02d:%02d' % (
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                )
            else:
                _svalue = '%02d:%02d:%02d.%s' % (
                    input_data.hour,
                    input_data.minute,
                    input_data.second,
                    ('%f' % (float(input_data.microsecond) / 1000000))[2:],
                )
            if input_data.tzinfo is not None:
                tzoff = input_data.tzinfo.utcoffset(input_data)
                if tzoff is not None:
                    total_seconds = tzoff.seconds + (86400 * tzoff.days)
                    if total_seconds == 0:
                        _svalue += 'Z'
                    else:
                        if total_seconds < 0:
                            _svalue += '-'
                            total_seconds *= -1
                        else:
                            _svalue += '+'
                        hours = total_seconds // 3600
                        minutes = (total_seconds - (hours * 3600)) // 60
                        _svalue += '{0:02d}:{1:02d}'.format(hours, minutes)
            return _svalue
        def gds_validate_simple_patterns(self, patterns, target):
            # pat is a list of lists of strings/patterns.  We should:
            # - AND the outer elements
            # - OR the inner elements
            found1 = True
            for patterns1 in patterns:
                found2 = False
                for patterns2 in patterns1:
                    if re_.search(patterns2, target) is not None:
                        found2 = True
                        break
                if not found2:
                    found1 = False
                    break
            return found1
        @classmethod
        def gds_parse_time(cls, input_data):
            tz = None
            if input_data[-1] == 'Z':
                tz = GeneratedsSuper._FixedOffsetTZ(0, 'UTC')
                input_data = input_data[:-1]
            else:
                results = GeneratedsSuper.tzoff_pattern.search(input_data)
                if results is not None:
                    tzoff_parts = results.group(2).split(':')
                    tzoff = int(tzoff_parts[0]) * 60 + int(tzoff_parts[1])
                    if results.group(1) == '-':
                        tzoff *= -1
                    tz = GeneratedsSuper._FixedOffsetTZ(
                        tzoff, results.group(0))
                    input_data = input_data[:-6]
            if len(input_data.split('.')) > 1:
                dt = datetime_.datetime.strptime(input_data, '%H:%M:%S.%f')
            else:
                dt = datetime_.datetime.strptime(input_data, '%H:%M:%S')
            dt = dt.replace(tzinfo=tz)
            return dt.time()
        def gds_str_lower(self, instring):
            return instring.lower()
        def get_path_(self, node):
            path_list = []
            self.get_path_list_(node, path_list)
            path_list.reverse()
            path = '/'.join(path_list)
            return path
        Tag_strip_pattern_ = re_.compile(r'\{.*\}')
        def get_path_list_(self, node, path_list):
            if node is None:
                return
            tag = GeneratedsSuper.Tag_strip_pattern_.sub('', node.tag)
            if tag:
                path_list.append(tag)
            self.get_path_list_(node.getparent(), path_list)
        def get_class_obj_(self, node, default_class=None):
            class_obj1 = default_class
            if 'xsi' in node.nsmap:
                classname = node.get('{%s}type' % node.nsmap['xsi'])
                if classname is not None:
                    names = classname.split(':')
                    if len(names) == 2:
                        classname = names[1]
                    class_obj2 = globals().get(classname)
                    if class_obj2 is not None:
                        class_obj1 = class_obj2
            return class_obj1
        def gds_build_any(self, node, type_name=None):
            return None
        @classmethod
        def gds_reverse_node_mapping(cls, mapping):
            return dict(((v, k) for k, v in mapping.items()))

    def getSubclassFromModule_(module, class_):
        '''Get the subclass of a class from a specific module.'''
        name = class_.__name__ + 'Sub'
        if hasattr(module, name):
            return getattr(module, name)
        else:
            return None


#
# If you have installed IPython you can uncomment and use the following.
# IPython is available from http://ipython.scipy.org/.
#

## from IPython.Shell import IPShellEmbed
## args = ''
## ipshell = IPShellEmbed(args,
##     banner = 'Dropping into IPython',
##     exit_msg = 'Leaving Interpreter, back to program.')

# Then use the following line where and when you want to drop into the
# IPython shell:
#    ipshell('<some message> -- Entering ipshell.\nHit Ctrl-D to exit')

#
# Globals
#

ExternalEncoding = 'utf-8'
Tag_pattern_ = re_.compile(r'({.*})?(.*)')
String_cleanup_pat_ = re_.compile(r"[\n\r\s]+")
Namespace_extract_pat_ = re_.compile(r'{(.*)}(.*)')
CDATA_pattern_ = re_.compile(r"<!\[CDATA\[.*?\]\]>", re_.DOTALL)

# Change this to redirect the generated superclass module to use a
# specific subclass module.
CurrentSubclassModule_ = None

#
# Support/utility functions.
#


def showIndent(outfile, level, pretty_print=True):
    if pretty_print:
        for idx in range(level):
            outfile.write('    ')


def quote_xml(inStr):
    "Escape markup chars, but do not modify CDATA sections."
    if not inStr:
        return ''
    s1 = (isinstance(inStr, string_types) and inStr or '%s' % inStr)
    s2 = ''
    pos = 0
    matchobjects = CDATA_pattern_.finditer(s1)
    for mo in matchobjects:
        s3 = s1[pos:mo.start()]
        s2 += quote_xml_aux(s3)
        s2 += s1[mo.start():mo.end()]
        pos = mo.end()
    s3 = s1[pos:]
    s2 += quote_xml_aux(s3)
    return s2


def quote_xml_aux(inStr):
    s1 = inStr.replace('&', '&amp;')
    s1 = s1.replace('<', '&lt;')
    s1 = s1.replace('>', '&gt;')
    return s1


def quote_attrib(inStr):
    s1 = (isinstance(inStr, string_types) and inStr or '%s' % inStr)
    s1 = s1.replace('&', '&amp;')
    s1 = s1.replace('<', '&lt;')
    s1 = s1.replace('>', '&gt;')
    if '"' in s1:
        if "'" in s1:
            s1 = '"%s"' % s1.replace('"', "&quot;")
        else:
            s1 = "'%s'" % s1
    else:
        s1 = '"%s"' % s1
    return s1


def quote_python(inStr):
    s1 = inStr
    if s1.find("'") == -1:
        if s1.find('\n') == -1:
            return "'%s'" % s1
        else:
            return "'''%s'''" % s1
    else:
        if s1.find('"') != -1:
            s1 = s1.replace('"', '\\"')
        if s1.find('\n') == -1:
            return '"%s"' % s1
        else:
            return '"""%s"""' % s1


def get_all_text_(node):
    if node.text is not None:
        text = node.text
    else:
        text = ''
    for child in node:
        if child.tail is not None:
            text += child.tail
    return text


def find_attr_value_(attr_name, node):
    attrs = node.attrib
    attr_parts = attr_name.split(':')
    value = None
    if len(attr_parts) == 1:
        value = attrs.get(attr_name)
    elif len(attr_parts) == 2:
        prefix, name = attr_parts
        namespace = node.nsmap.get(prefix)
        if namespace is not None:
            value = attrs.get('{%s}%s' % (namespace, name, ))
    return value


class GDSParseError(Exception):
    pass


def raise_parse_error(node, msg):
    msg = '%s (element %s/line %d)' % (msg, node.tag, node.sourceline, )
    raise GDSParseError(msg)


class MixedContainer:
    # Constants for category:
    CategoryNone = 0
    CategoryText = 1
    CategorySimple = 2
    CategoryComplex = 3
    # Constants for content_type:
    TypeNone = 0
    TypeText = 1
    TypeString = 2
    TypeInteger = 3
    TypeFloat = 4
    TypeDecimal = 5
    TypeDouble = 6
    TypeBoolean = 7
    TypeBase64 = 8
    def __init__(self, category, content_type, name, value):
        self.category = category
        self.content_type = content_type
        self.name = name
        self.value = value
    def getCategory(self):
        return self.category
    def getContenttype(self, content_type):
        return self.content_type
    def getValue(self):
        return self.value
    def getName(self):
        return self.name
    def export(self, outfile, level, name, namespace, pretty_print=True):
        if self.category == MixedContainer.CategoryText:
            # Prevent exporting empty content as empty lines.
            if self.value.strip():
                outfile.write(self.value)
        elif self.category == MixedContainer.CategorySimple:
            self.exportSimple(outfile, level, name)
        else:    # category == MixedContainer.CategoryComplex
            self.value.export(outfile, level, namespace, name, pretty_print)
    def exportSimple(self, outfile, level, name):
        if self.content_type == MixedContainer.TypeString:
            outfile.write('<%s>%s</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeInteger or \
                self.content_type == MixedContainer.TypeBoolean:
            outfile.write('<%s>%d</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeFloat or \
                self.content_type == MixedContainer.TypeDecimal:
            outfile.write('<%s>%f</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeDouble:
            outfile.write('<%s>%g</%s>' % (
                self.name, self.value, self.name))
        elif self.content_type == MixedContainer.TypeBase64:
            outfile.write('<%s>%s</%s>' % (
                self.name, base64.b64encode(self.value), self.name))
    def to_etree(self, element):
        if self.category == MixedContainer.CategoryText:
            # Prevent exporting empty content as empty lines.
            if self.value.strip():
                if len(element) > 0:
                    if element[-1].tail is None:
                        element[-1].tail = self.value
                    else:
                        element[-1].tail += self.value
                else:
                    if element.text is None:
                        element.text = self.value
                    else:
                        element.text += self.value
        elif self.category == MixedContainer.CategorySimple:
            subelement = etree_.SubElement(element, '%s' % self.name)
            subelement.text = self.to_etree_simple()
        else:    # category == MixedContainer.CategoryComplex
            self.value.to_etree(element)
    def to_etree_simple(self):
        if self.content_type == MixedContainer.TypeString:
            text = self.value
        elif (self.content_type == MixedContainer.TypeInteger or
                self.content_type == MixedContainer.TypeBoolean):
            text = '%d' % self.value
        elif (self.content_type == MixedContainer.TypeFloat or
                self.content_type == MixedContainer.TypeDecimal):
            text = '%f' % self.value
        elif self.content_type == MixedContainer.TypeDouble:
            text = '%g' % self.value
        elif self.content_type == MixedContainer.TypeBase64:
            text = '%s' % base64.b64encode(self.value)
        return text
    def exportLiteral(self, outfile, level, name):
        if self.category == MixedContainer.CategoryText:
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s", "%s"),\n' % (
                    self.category, self.content_type, self.name, self.value))
        elif self.category == MixedContainer.CategorySimple:
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s", "%s"),\n' % (
                    self.category, self.content_type, self.name, self.value))
        else:    # category == MixedContainer.CategoryComplex
            showIndent(outfile, level)
            outfile.write(
                'model_.MixedContainer(%d, %d, "%s",\n' % (
                    self.category, self.content_type, self.name,))
            self.value.exportLiteral(outfile, level + 1)
            showIndent(outfile, level)
            outfile.write(')\n')


class MemberSpec_(object):
    def __init__(self, name='', data_type='', container=0):
        self.name = name
        self.data_type = data_type
        self.container = container
    def set_name(self, name): self.name = name
    def get_name(self): return self.name
    def set_data_type(self, data_type): self.data_type = data_type
    def get_data_type_chain(self): return self.data_type
    def get_data_type(self):
        if isinstance(self.data_type, list):
            if len(self.data_type) > 0:
                return self.data_type[-1]
            else:
                return 'xs:string'
        else:
            return self.data_type
    def set_container(self, container): self.container = container
    def get_container(self): return self.container


def _cast(typ, value):
    if typ is None or value is None:
        return value
    return typ(value)

#
# Data representation classes.
#


class diffExportType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, dbState=None, diffState=None, deleted=None, added=None):
        self.original_tagname_ = None
        self.dbState = _cast(None, dbState)
        self.diffState = _cast(None, diffState)
        if deleted is None:
            self.deleted = []
        else:
            self.deleted = deleted
        if added is None:
            self.added = []
        else:
            self.added = added
    def factory(*args_, **kwargs_):
        if CurrentSubclassModule_ is not None:
            subclass = getSubclassFromModule_(
                CurrentSubclassModule_, diffExportType)
            if subclass is not None:
                return subclass(*args_, **kwargs_)
        if diffExportType.subclass:
            return diffExportType.subclass(*args_, **kwargs_)
        else:
            return diffExportType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_deleted(self): return self.deleted
    def set_deleted(self, deleted): self.deleted = deleted
    def add_deleted(self, value): self.deleted.append(value)
    def insert_deleted_at(self, index, value): self.deleted.insert(index, value)
    def replace_deleted_at(self, index, value): self.deleted[index] = value
    def get_added(self): return self.added
    def set_added(self, added): self.added = added
    def add_added(self, value): self.added.append(value)
    def insert_added_at(self, index, value): self.added.insert(index, value)
    def replace_added_at(self, index, value): self.added[index] = value
    def get_dbState(self): return self.dbState
    def set_dbState(self, dbState): self.dbState = dbState
    def get_diffState(self): return self.diffState
    def set_diffState(self, diffState): self.diffState = diffState
    def hasContent_(self):
        if (
            self.deleted or
            self.added
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='diffexport:', name_='diffExportType', namespacedef_='xmlns:diffexport="http://ebas.nilu.no/EBAS_diffExport/V1.0"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='diffExportType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='diffexport:', name_='diffExportType', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='diffexport:', name_='diffExportType'):
        if self.dbState is not None and 'dbState' not in already_processed:
            already_processed.add('dbState')
            outfile.write(' dbState=%s' % (quote_attrib(self.dbState), ))
        if self.diffState is not None and 'diffState' not in already_processed:
            already_processed.add('diffState')
            outfile.write(' diffState=%s' % (quote_attrib(self.diffState), ))
    def exportChildren(self, outfile, level, namespace_='diffexport:', name_='diffExportType', fromsubclass_=False, pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        for deleted_ in self.deleted:
            deleted_.export(outfile, level, namespace_, name_='deleted', pretty_print=pretty_print)
        for added_ in self.added:
            added_.export(outfile, level, namespace_, name_='added', pretty_print=pretty_print)
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('dbState', node)
        if value is not None and 'dbState' not in already_processed:
            already_processed.add('dbState')
            self.dbState = value
        value = find_attr_value_('diffState', node)
        if value is not None and 'diffState' not in already_processed:
            already_processed.add('diffState')
            self.diffState = value
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'deleted':
            obj_ = dataIntervalType.factory()
            obj_.build(child_)
            self.deleted.append(obj_)
            obj_.original_tagname_ = 'deleted'
        elif nodeName_ == 'added':
            obj_ = dataIntervalType.factory()
            obj_.build(child_)
            self.added.append(obj_)
            obj_.original_tagname_ = 'added'
# end class diffExportType


class dataIntervalType(GeneratedsSuper):
    subclass = None
    superclass = None
    def __init__(self, datasetKey=None, startTime=None, endTime=None):
        self.original_tagname_ = None
        self.datasetKey = _cast(int, datasetKey)
        self.startTime = _cast(None, startTime)
        self.endTime = _cast(None, endTime)
    def factory(*args_, **kwargs_):
        if CurrentSubclassModule_ is not None:
            subclass = getSubclassFromModule_(
                CurrentSubclassModule_, dataIntervalType)
            if subclass is not None:
                return subclass(*args_, **kwargs_)
        if dataIntervalType.subclass:
            return dataIntervalType.subclass(*args_, **kwargs_)
        else:
            return dataIntervalType(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_datasetKey(self): return self.datasetKey
    def set_datasetKey(self, datasetKey): self.datasetKey = datasetKey
    def get_startTime(self): return self.startTime
    def set_startTime(self, startTime): self.startTime = startTime
    def get_endTime(self): return self.endTime
    def set_endTime(self, endTime): self.endTime = endTime
    def hasContent_(self):
        if (

        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='diffexport:', name_='dataIntervalType', namespacedef_='xmlns:diffexport="http://ebas.nilu.no/EBAS_diffExport/V1.0"', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='dataIntervalType')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='diffexport:', name_='dataIntervalType', pretty_print=pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='diffexport:', name_='dataIntervalType'):
        if self.datasetKey is not None and 'datasetKey' not in already_processed:
            already_processed.add('datasetKey')
            outfile.write(' datasetKey="%s"' % self.gds_format_integer(self.datasetKey, input_name='datasetKey'))
        if self.startTime is not None and 'startTime' not in already_processed:
            already_processed.add('startTime')
            outfile.write(' startTime=%s' % (quote_attrib(self.startTime), ))
        if self.endTime is not None and 'endTime' not in already_processed:
            already_processed.add('endTime')
            outfile.write(' endTime=%s' % (quote_attrib(self.endTime), ))
    def exportChildren(self, outfile, level, namespace_='diffexport:', name_='dataIntervalType', fromsubclass_=False, pretty_print=True):
        pass
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('datasetKey', node)
        if value is not None and 'datasetKey' not in already_processed:
            already_processed.add('datasetKey')
            try:
                self.datasetKey = int(value)
            except ValueError as exp:
                raise_parse_error(node, 'Bad integer attribute: %s' % exp)
        value = find_attr_value_('startTime', node)
        if value is not None and 'startTime' not in already_processed:
            already_processed.add('startTime')
            self.startTime = value
        value = find_attr_value_('endTime', node)
        if value is not None and 'endTime' not in already_processed:
            already_processed.add('endTime')
            self.endTime = value
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        pass
# end class dataIntervalType


GDSClassesMapping = {
    'added': dataIntervalType,
    'deleted': dataIntervalType,
    'diffExport': diffExportType,
}


USAGE_TEXT = """
Usage: python <Parser>.py [ -s ] <in_xml_file>
"""


def usage():
    print(USAGE_TEXT)
    sys.exit(1)


def get_root_tag(node):
    tag = Tag_pattern_.match(node.tag).groups()[-1]
    rootClass = GDSClassesMapping.get(tag)
    if rootClass is None:
        rootClass = globals().get(tag)
    return tag, rootClass


def parse(inFileName, silence=False):
    parser = None
    doc = parsexml_(inFileName, parser)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'diffExportType'
        rootClass = diffExportType
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('<?xml version="1.0" ?>\n')
        rootObj.export(
            sys.stdout, 0, name_=rootTag,
            namespacedef_='xmlns:diffexport="http://ebas.nilu.no/EBAS_diffExport/V1.0"',
            pretty_print=True)
    return rootObj


def parseEtree(inFileName, silence=False):
    parser = None
    doc = parsexml_(inFileName, parser)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'diffExportType'
        rootClass = diffExportType
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    mapping = {}
    rootElement = rootObj.to_etree(None, name_=rootTag, mapping_=mapping)
    reverse_mapping = rootObj.gds_reverse_node_mapping(mapping)
    if not silence:
        content = etree_.tostring(
            rootElement, pretty_print=True,
            xml_declaration=True, encoding="utf-8")
        sys.stdout.write(content)
        sys.stdout.write('\n')
    return rootObj, rootElement, mapping, reverse_mapping


def parseString(inString, silence=False):
    from six import StringIO
    parser = None
    doc = parsexml_(StringIO(inString), parser)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'diffExportType'
        rootClass = diffExportType
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('<?xml version="1.0" ?>\n')
        rootObj.export(
            sys.stdout, 0, name_=rootTag,
            namespacedef_='xmlns:diffexport="http://ebas.nilu.no/EBAS_diffExport/V1.0"')
    return rootObj


def parseLiteral(inFileName, silence=False):
    parser = None
    doc = parsexml_(inFileName, parser)
    rootNode = doc.getroot()
    rootTag, rootClass = get_root_tag(rootNode)
    if rootClass is None:
        rootTag = 'diffExportType'
        rootClass = diffExportType
    rootObj = rootClass.factory()
    rootObj.build(rootNode)
    # Enable Python to collect the space used by the DOM.
    doc = None
    if not silence:
        sys.stdout.write('#from ebas_diffexport_generated import *\n\n')
        sys.stdout.write('import ebas_diffexport_generated as model_\n\n')
        sys.stdout.write('rootObj = model_.rootClass(\n')
        rootObj.exportLiteral(sys.stdout, 0, name_=rootTag)
        sys.stdout.write(')\n')
    return rootObj


def main():
    args = sys.argv[1:]
    if len(args) == 1:
        parse(args[0])
    else:
        usage()


if __name__ == '__main__':
    #import pdb; pdb.set_trace()
    main()


__all__ = [
    "dataIntervalType",
    "diffExportType"
]



"""
$Id: cy.py 2466 2020-07-10 13:20:48Z pe $

EBAS Masterdata Class for component synonym masterdata

This module implements the class EbasMasterCY.

"""

from .. import EbasDomainError
from .offline_masterdata import CYOfflineMasterData
from .co import EbasMasterCO
from .base import EbasMasterBase

class EbasMasterCY(EbasMasterBase, CYOfflineMasterData):
    """
    Domain Class for component synonym masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling components and checking them against master data.
    Components master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for CO (components)
    # Those are fallback values, will be read from database as soon as possible.
    CYOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read co masterdata if dbh is provided.
        """
        CYOfflineMasterData.__init__(self)
        self._co = EbasMasterCO(dbh)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)

    def lookup_synonym_strict(self, name, case_sensitive=True,
                              only_historic=False):
        """
        Lookup a strict synonym.
        Try CO_COMP_NAME first, then try strict synonyms, at last try case
        insensitive Lookup Name search (strict).
        Parameters:
            name     string with component name that should be identified
        Returns:
            CO_COMP_NAME (must always be max hit when using strict synonyms!)
        Raises:
            DoaminError if not found
            RuntimeError when strictness is inconsistent
        """
        try:
            return self._co[name].CO_COMP_NAME
        except KeyError:
            pass

        # lookup case sensitive
        hits = [x for x in list(self.__class__.META.keys())
                if any(
                    [True for y in self.__class__.META[x]
                     if y['CY_STRICT'] == 1 and y['CY_SYNONYM'] == name and
                        (only_historic == False or \
                             y['CN_SCHEMA'] == 'Historic EBAS Name')])]
        if len(hits) == 1:
            return hits[0]
        if len(hits) > 1:
            raise RuntimeError()

        if case_sensitive == False:
            # lookup case insensitive
            hits = [x for x in list(self.__class__.META.keys())
                    if any(
                        [True for y in self.__class__.META[x]
                         if y['CY_STRICT'] == 1 and \
                            y['CN_SCHEMA'] == 'Lookup Name' and \
                            y['CY_SYNONYM'].lower() == name.lower()])]
            # special cases: case insensitive lookup led to multiple hits
            # e.g.: co and CO
            if len(hits) > 1:
                amb = ["{} ({})".format(y['CY_SYNONYM'], y['CO_COMP_NAME'])
                       for x in list(self.__class__.META.values()) for y in x
                       if y['CY_STRICT'] == 1 and \
                          y['CN_SCHEMA'] == 'Lookup Name' and \
                          y['CY_SYNONYM'].lower() == name.lower()]
                raise EbasDomainError(
                    "component lookup name '{}' is ambiguous for case insensitive "
                    "lookup - may be {}".format(name, ', '.join(amb)))
            if len(hits) == 1:
                return hits[0]
        raise EbasDomainError("unknown component '{}'".format(name))

    @staticmethod
    def get_schema_ordered(cy_list):
        """
        Get an ordered list of elements with schema.
        """
        order = ('CAS Number', 'IUPAC Name', 'Historic EBAS Name',
                 'Molecular Formula', 'Condensed Formula', 'Lookup Name')
        retlist = []
        for elem in order:
            synlist = []
            for syn in cy_list:
                # there may be actually more then one synonym for one schema and
                # one component...
                # so make a list and sort it afterwards
                if syn['CN_SCHEMA'] == elem:
                    synlist.append(syn)
            retlist += sorted(synlist, key=lambda x: x['CY_SYNONYM'])
        synlist = []
        for syn in cy_list:
            # other synonyms with schema: sort by schama, synonym
            if syn['CN_SCHEMA'] is not None and syn['CN_SCHEMA'] not in order:
                synlist.append(syn)
        retlist += sorted(synlist,
                          key=lambda x: (x['CN_SCHAMA'], x['CY_SYNONYM']))
        return retlist

    @staticmethod
    def get_noschema_ordered(cy_list):
        """
        Get an ordered list of elements without schema.
        """
        return sorted([x for x in cy_list if x['CN_SCHEMA'] is None],
                      key=lambda x: x['CY_SYNONYM'])

    @staticmethod
    def get_all_ordered(cy_list):
        """
        Get an ordered list of elements without schema.
        """
        return EbasMasterCY.get_schema_ordered(cy_list) + \
            EbasMasterCY.get_noschema_ordered(cy_list)



"""
nilu/datetime_helper.py
$Id: datetime_helper.py 2686 2021-08-18 11:24:59Z pe $

Module for helper functions for date/time handling.
Provides classes for DateTime intervals and sets of intervals.

History:
V.1.0.0  2012-10-05  pe  initial version
V.2.0.0  2016-12-08  pe  rewritten, created classes

"""

import datetime
import re
import math
from dateutil.relativedelta import relativedelta

class DatetimeInterval(list):
    """
    Time interval class.

    Description of time intervals:
     - Generally, an interval has a start and an end time.
     - Start or End can be None which means + and - infinity, e.g.
       [None, 2014-01-01] is all times before and including
       2014-01-01T00:00:00.000.
     - None is interpreted as empty interval []
     - intervals must be positive (start<end for () and [) intervals,
       start<=end for [] intervals)
    """

    def __init__(self, start, end, empty=False, bounds='[)'):
        """
        Initialize a time interval
        """
        self.bounds = bounds
        if empty:
            # initialize the list:
            super(DatetimeInterval, self).__init__([])
            self.bounds = ''
            return
        # initialize the list:
        super(DatetimeInterval, self).__init__((start, end))
        # ignore and hard reset open bounds for infinity:
        if start is None:
            self.bounds = '(' + self.bounds[1]
        if end is None:
            self.bounds = self.bounds[0] + ')'
        self.check()

    def __str__(self):
        """
        Create a pretty string.
        """
        if self.is_empty():
            return self.bounds
        ret = ['-Inf', 'Inf']
        if self[0]:
            ret[0] = datetime_string(self[0])
        if self[1]:
            ret[1] = datetime_string(self[1])
        return "{}{}, {}{}".format(self.bounds[0], ret[0], ret[1], self.bounds[1])

    def __repr__(self):
        """
        Use the same es str?
        """
        return str(self)

    def __eq__(self, other):
        """
        Check for equality.
        """
        return self.cmp(other) == 0
        
    def __ne__(self, other):
        """
        Check for inequality.
        """
        return self.cmp(other) != 0
        
    def __lt__(self, other):
        """
        Check if less than.
        """
        return self.cmp(other) < 0
        
    def __le__(self, other):
        """
        Check if less or equal.
        """
        return self.cmp(other) <= 0
        
    def __gt__(self, other):
        """
        Check if greater than.
        """
        return self.cmp(other) > 0
        
    def __ge__(self, other):
        """
        Check if greater or equal.
        """
        return self.cmp(other) >= 0

    def __hash__(self):
        """
        Hash value for object
        """
        return hash(tuple(self) + (self.bounds,))

    def cmp(self, other):
        """
        Compare two objects
        """
        if self.is_empty() and other.is_empty():
            return 0
        if self.is_empty():
            return -1
        if other.is_empty():
            return 1

        if (self[0] is None and other[0] is not None) or \
                (self[0] is not None and other[0] is not None and \
                 self[0] < other[0]):
            return -1
        if (other[0] is None and self[0] is not None) or \
                (self[0] is not None and other[0] is not None and \
                 self[0] > other[0]):
            return 1
        # here the start values are equal (either both None or both the same date)
        if self.bounds[0] == '[' and other.bounds[0] == '(':
            return -1
        if self.bounds[0] == '(' and other.bounds[0] == ']':
            return 1
        # here they have equal start (also bound)
        if (self[1] is None and other[1] is not None) or \
                (self[1] is not None and other[1] is not None and \
                 self[1] > other[1]):
            return 1
        if (other[1] is None and self[1] is not None) or \
                (self[1] is not None and other[1] is not None and \
                 self[1] < other[1]):
            return -1
        # end values are also equal
        if self.bounds[1] == ')' and other.bounds[1] == ']':
            return -1
        if self.bounds[1] == ']' and other.bounds[1] == ')':
            return 1
        return 0

    @classmethod
    def empty(cls):
        """
        Create an empty interval.
        """
        return DatetimeInterval(None, None, empty=True)

    def is_empty(self):
        """
        An empty interval is one with zero length!
        """
        return not bool(self)

    def is_unbound(self):
        """
        Is unbound at both ends.
        """
        if not self:
            return False
        if self[0] is None and self[1] is None:
            return True
        return False

    def clone(self):
        """
        Clones the interval. Factory of DatetimeInterval.
        Parameters:
            None
        Returns:
            copy of self
        """
        if self.is_empty():
            return self.empty()
        return DatetimeInterval(*self, bounds=self.bounds)

    def check(self):
        """
        Check an interval.
        Parameters:
            interval    time interval to be checked
        Returns:
            None
        Raises:
            TypeError, ValueError
        """
        if self.is_empty():
            return
        if self.bounds not in ('[]', '[)', '(]', '()'):
            raise ValueError('illegal interval bounds {}'.format(self.bounds))
        intv = list(self)  # raises TypeError if it cannot be converted
        if len(intv) != 2:
            raise ValueError('illegal time interval')
        for tim in intv:
            if tim is not None and not isinstance(tim, datetime.datetime):
                raise TypeError('time interval limits must be None or '
                                'datetime.datetime')
        if intv[0] is not None and intv[1] is not None:
            if (self.bounds[0] == '(' or self.bounds[1] == ')') and \
                    intv[0] >= intv[1]:
                raise ValueError(
                    'time interval start must be earlier then time interval '
                    'end for open and half-open intervals')
            elif intv[0] > intv[1]:
                raise ValueError(
                    'time interval start must be earlier or equal time '
                    'interval end for closed intervals')

    def overlap(self, interval2):
        """
        Checks intervals for overlap.
        Returns overlaping time interval (intersection).

        Parameters:
            interval2   interval to overlap self wdith
        Returns:
            DatetimeInterval (overlapping interval)
            None if no overlap
        """
        if self.is_empty() or interval2.is_empty():
            # empty intervals cannot overlap with anything
            return DatetimeInterval.empty()
        if (not self[0] or not interval2[1] or \
            (self.bounds[0] == '[' and interval2.bounds[1] == ']' and \
             self[0] <= interval2[1]) or \
            self[0] < interval2[1]) and \
           (not self[1] or not interval2[0] or \
            (self.bounds[1] == ']' and interval2.bounds[0] == '[' and \
             self[1] >= interval2[0]) or \
            self[1] > interval2[0]):
            if self[0] is None:
                if interval2[0] is None:
                    low = None
                    lowbnd = '('
                else:
                    low = interval2[0]
                    lowbnd = interval2.bounds[0]
            elif interval2[0] is None:
                low = self[0]
                lowbnd = self.bounds[0]
            elif self[0] > interval2[0]:
                low = self[0]
                lowbnd = self.bounds[0]
            elif self[0] < interval2[0]:
                low = interval2[0]
                lowbnd = interval2.bounds[0]
            else:
                # equal
                low = self[0]
                lowbnd = self.bounds[0] if interval2.bounds[0] == '[' else '('

            if self[1] is None:
                if interval2[1] is None:
                    upp = None
                    uppbnd = ')'
                else:
                    upp = interval2[1]
                    uppbnd = interval2.bounds[1]
            elif interval2[1] is None:
                upp = self[1]
                uppbnd = self.bounds[1]
            elif self[1] < interval2[1]:
                upp = self[1]
                uppbnd = self.bounds[1]
            elif self[1] > interval2[1]:
                upp = interval2[1]
                uppbnd = interval2.bounds[1]
            else:
                # equal
                upp = interval2[1]
                uppbnd = self.bounds[1] if interval2.bounds == ']' else ')'
            return DatetimeInterval(low, upp, bounds=lowbnd+uppbnd)
        else:
            return DatetimeInterval.empty()

    def contains_timestamp(self, dtime):
        """
        Checks if timestamp is contained in interval.
        Parameters:
            dtime    datetime.datetime (contained in me?)
                     if None we assume unlimited end of interval.
        Returns:
            bool
        """
        if self.is_empty():
            return False
        if ((dtime is None and self[1] is None) or
            (dtime is not None and
             (self[0] is None or self[0] <= dtime) and
             (self[1] is None or self[1] > dtime))):
            return True
        return False

    def contains(self, other):
        """
        Checks if other interval is contained in interval.
        (Fully contained, otherwise it would be an overlap).
        Parameters:
            other    other interval (contained in me?)
        Returns:
            bool
        """
        if self.is_empty() or other.is_empty():
            # empty intervals cannot contain or be contained
            return False
        # check lower limit: if...
        #  both are numbers and other not fully in self OR
        #  both are numbers and equal and
        #      self is open and other is closed OR
        #  other[0] is infinite and self[0] is not infinite OR
        #  other[0] is infinite and self[0] is infinite and
        #      self is open and other is closed
        if self[0] and other[0] and self[0] > other[0] or \
           self[0] and other[0] and self[0] == other[0] and \
               self.bounds[0] == '(' and other.bounds == '[' or \
           not other[0] and self[0] or \
           not other[0] and not self[0] and \
               self.bounds[0] == '(' and other.bounds == '[':
            return False
        # check upper limit: if...
        #  both are numbers and other not fully in self OR
        #  both are numbers and equal and
        #      self is open and other is closed OR
        #  other[1] is infinite and self[1] is not infinite OR
        #  other[1] is infinite and self[1] is infinite and
        #      self is open and other is closed
        if self[1] and other[1] and self[1] < other[1] or \
           self[1] and other[1] and self[1] == other[1] and \
               self.bounds[1] == ')' and other.bounds == ']' or \
           not other[1] and self[1] or \
           not other[1] and not self[1] and \
               self.bounds[1] == ')' and other.bounds == ']':
            return False
        # else, other is contained in self
        return True

    def distance(self, interval2):
        """
        Calculates the distance (in seconds).
        0.0 is either overlap (contains) or adjoins.
        """
        if self.overlap(interval2) or self.adjoins(interval2):
            return 0.0
        if self[0] is not None and interval2[1] is not None and (self[0]-interval2[1]).total_seconds() > 0.0:
            return (self[0]-interval2[1]).total_seconds()
        if self[1] is not None and interval2[0] is not None and (interval2[0]-self[1]).total_seconds() > 0.0:
            return (interval2[0]-self[1]).total_seconds()

    def subtract(self, interval2):
        # pylint: disable=R0912, W1401
        #  R0912: Too many branches
        """
        Subtracts interval2 from self.
        (in set theory: relative complement of interval2 in self:
        self \\ interval2)

        Parameters:
            interval2   interval to subtract from self
        Returns:
            DatetimeIntervalSet   resulting interval(s) (can be empty)
        """
        # pylint: enable=W1401
        ret = DatetimeIntervalSet()
        if self.is_empty():
            # nothing to subtract from
            return ret
        if not self.overlap(interval2):
            # nothing to subtract
            ret.add(self.clone())
            return ret
        if (self[0] is None and (interval2[0] is not None or
                                 (self.bounds[0] == '[' and
                                  interval2.bounds[0] == '('))) or \
           (self[0] is not None and interval2[0] is not None and
            (self[0] < interval2[0] or (self[0] == interval2[0] and
                                        self.bounds[0] == '[' and
                                        interval2.bounds[0] == '('))):
            # There is a part extending at the beginning of self
            ret.add(DatetimeInterval(
                self[0], interval2[0],
                bounds=self.bounds[0] + (']' if interval2.bounds[0] == '('
                                         else ')')))

        if (self[1] is None and (interval2[1] is not None or
                                 (self.bounds[1] == ']' and 
                                  interval2.bounds[1] == ')'))) or \
           (self[1] is not None and interval2[1] is not None and
            (self[1] > interval2[1] or (self[1] == interval2[1] and
                                        self.bounds[1] == ']' and
                                        interval2.bounds[1] == ')'))):
            # There is something extending at the end of self
            ret.add(DatetimeInterval(
                interval2[1], self[1],
                bounds=('[' if interval2.bounds[1] == ')' else '(') + \
                    self.bounds[1]))        
        return ret

    def adjoins_right(self, interval2):
        """
        Checks if interval2 adjoins self to the right.
        Parameters:
            interval2   second interval
        Returns:
            True / False
        """
        if self.is_empty() or interval2 is None or interval2.is_empty():
            return False  # empty intervals can not adjoin to anything
        return bool(self[1] == interval2[0] and
                    ((self.bounds[1] == ')' and interval2.bounds[0] == '[') or
                     (self.bounds[1] == ']' and interval2.bounds[0] == '(')))

    def adjoins_left(self, interval2):
        """
        Checks if interval2 adjoins self to the left.
        Parameters:
            interval2   second interval
        Returns:
            True / False
        """
        if self.is_empty() or interval2 is None or interval2.is_empty():
            return False  # empty intervals can not adjoin to anything
        return bool(self[0] == interval2[1] and
                    ((self.bounds[0] == '[' and interval2.bounds[1] == ')') or
                     (self.bounds[0] == '(' and interval2.bounds[1] == ']')))

    def adjoins(self, interval2):
        """
        Checks if interval2 adjoins self to the left or right.
        Parameters:
            interval2   second interval
        Returns:
            True / False
        """
        return self.adjoins_left(interval2) or self.adjoins_right(interval2)

class DatetimeIntervalSet(list):
    """
    Set of DateTime Intervals.
    """

    def __init__(self, initintv=None):
        """
        Initialize a time interval
        Parameters:
            initintv   initial interval to be added righrt away
        """
        # initialize the list:
        super(DatetimeIntervalSet, self).__init__([])
        if initintv:
            self.add(initintv)

    def clone(self):
        """
        Clones the interval set.
        Parameters:
            None
        Returns:
            cloned object
        """
        ret =DatetimeIntervalSet()
        for intv in self:
            ret.add(intv.clone())
        return ret

    def union(self, interval):
        """
        Calculate the union of interval set with an interval or another interval
        set.
        Like add, but returns the union instead of changing the object.

        Parameters:
            interval    DatetimeInterval or DatetimeIntervalSet to be unioned
        Returns:
            union (DatetimeIntervalSet)
        """
        ret = self.clone()
        ret.add(interval)
        return ret
        
    def add(self, interval):
        # pylint: disable=R0912
        #  R0912: Too many branches
        """
        Adds interval or interval set to the interval set.
        Like union, but changes the object instead of returning the result.

        Parameters:
            interval    DatetimeInterval or DatetimeIntervalSet to be added
        Returns:
            None
        """
        if isinstance(interval, DatetimeInterval):
            self._add_interval(interval)
        else:
            for intv in interval:
                self._add_interval(intv)

    def _add_interval(self, interval):
        # pylint: disable=R0912
        #  R0912: Too many branches
        """
        Adds interval2 to the interval set.

        Parameters:
            interval    DatetimeInterval to be added
        Returns:
            None
        """
        if not interval:
            return  # empty interval, nothing to add
        added = False
        i = 0
        while i < len(self):
            # check overlap/adjoin with interval
            if (self[i][0] is None or interval[1] is None or \
                self[i][0] <= interval[1]) and \
               (self[i][1] is None or interval[0] is None or \
                self[i][1] >= interval[0]):
                # element i overlaps with or adjoins interval: merge them:
                if self[i][0] is None or interval[0] is None:
                    self[i][0] = None
                else:
                    self[i][0] = min(self[i][0], interval[0])
                if self[i][1] is None or interval[1] is None:
                    self[i][1] = None
                else:
                    self[i][1] = max(self[i][1], interval[1])
                added = True
            # check overlap/adjoin with any other list element:
            j = i + 1
            while j < len(self):
                if (self[i][0] is None or self[j][1] is None or \
                    self[i][0] <= self[j][1]) and \
                   (self[i][1] is None or self[j][0] is None or \
                    self[i][1] >= self[j][0]):
                    # element i overlaps or adjoins element j:
                    # merge them and del j
                    if self[i][0] is None or self[j][0] is None:
                        self[i][0] = None
                    else:
                        self[i][0] = min(self[i][0], self[j][0])
                    if self[i][1] is None or self[j][1] is None:
                        self[i][1] = None
                    else:
                        self[i][1] = max(self[i][1], self[j][1])
                    del self[j]
                    # do not increment
                else:
                    j += 1
            i += 1
        if not added:
            self.append(interval.clone())
        self.sort(
            key=lambda x: [datetime.datetime.min if x[0] is None else x[0],
                           0 if x.bounds[0] == '[' else 1,
                           datetime.datetime.max if x[1] is None else x[1],
                           0 if x.bounds[0] == ')' else 1])

    def difference(self, interval):
        """
        Calculates the (set) difference of an interval set and an interval or
        another interval set.
        Like subtract, but returns the difference instead of changing the
        object.

        Parameters:
            interval  DatetimeInterval or DatetimeIntervalSet to be differenced
        Returns:
            None
        """
        if isinstance(interval, DatetimeInterval):
            return self._difference_interval(interval)
        else:
            ret = None
            for intv in interval:
                if not ret:
                    ret = self._difference_interval(intv)
                else:
                    ret = ret._difference_interval(intv)
                if not ret:
                    return ret

    def _difference_interval(self, interval):
        """
        Calculates the (set) difference of an interval set and an interval.

        Parameters:
            interval    DatetimeInterval to be differenced
        Returns:
            difference (DattimeIntervalSet)
        """
        ret= DatetimeIntervalSet()
        for intv in self:
            remain = intv.subtract(interval)
            if remain:
                ret.extend(remain)
        ret.sort()
        return ret

    def subtract(self, interval):
        """
        Subtracts an interval from the interval set.
        Like difference, but changes the object instead of returning the result.

        Parameters:
            interval    DatetimeInterval or DatetimeIntervalSet to be subtracted
        Returns:
            None
        """
        new = self.difference(interval)
        del self[:]
        self.extend(new)

    def intersection(self, interval):
        """
        Calculates the intersection of an interval set and and interval or
        another interval set.

        Parameters:
            interval   DatetimeInterval or DatetimeIntervalSet to be intersected
        Returns:
            None
        """
        ret = DatetimeIntervalSet()
        if isinstance(interval, DatetimeInterval):
            interval = [interval]
        for int1 in interval:
            for int2 in self:
                overlap = int1.overlap(int2)
                if overlap:
                    ret.add(overlap)
        return ret

    def overlap(self, interval):
        """
        Overlaps the interval set with an interval or another interval set.
        Like intersection, but changes the object instead of returning the
        result.

        Parameters:
            interval    DatetimeInterval or DatetimeIntervalSet to be overlapped
        Returns:
            None
        """
        new = self.intersection(interval)
        del self[:]
        self.extend(new)

def datetime_string(dtt):
    """
    Convert datetime object to a pretty string. Use precision according to
    content.
    Parameters:
        dtt    datetime object
    Returns:
        stting ipretty formatted datetime or "None"
    """
    if not dtt:
        return "None"
    fmt = '%Y-%m-%d'
    if dtt.microsecond:
        fmt += 'T%H:%M:%S.%f'
    elif dtt.second:
        fmt += 'T%H:%M:%S'
    elif dtt.minute or dtt.hour:
        fmt += 'T%H:%M'
    return dtt.strftime(fmt)

def datetime_round(datetime_, sec_digits):
    """
    Rounds a datatime.datetime object.
    Parameters:
        datetime_    datetime.datetime object
        sec_digits   digits of a second:
                       e.g.: 3: round to milliseconds
                       e.g.: 0: round to seconds
    Returns:
        datetime.datetime (must return new value, datetime is immutable)
    """
    seconds = datetime_.second + datetime_.microsecond / 1000000.0
    rounded = round(seconds, sec_digits)
    return datetime_ + datetime.timedelta(seconds=rounded-seconds)

def datetime_parse_iso8601(string):
    """
    Parses a single time string and returns a datetime.datetime object.

    Parameters:
        string    time string
    Returns:
        datetime
    Raises:
        ValueError in case of parsing errors

    TODO: support for optional timezones, week number syntax, ordinal day syntax
    """
    # check base ISO format syntax
    if not re.match('[0-9][0-9][0-9][0-9](-[0-9][0-9](-[0-9][0-9]' + \
       r'(T[0-9][0-9](:[0-9][0-9](:[0-9][0-9](\.[0-9]+)?)?)?)?)?)?$', string):
        raise ValueError('illegal ISO8601 time: ' + string)

    if len(string) > 20:
        fmt = "%Y-%m-%dT%H:%M:%S.%f"
    elif len(string) < 4:
        fmt = '%Y'
    else:
        fmt = "%Y-%m-%dT%H:%M:%S."[0:len(string)-2]
    try:
        return datetime.datetime.strptime(string, fmt)
    except ValueError:
        raise ValueError('illegal ISO8601 time: ' + string)

class DatetimeISO8601Duration(object):
    """
    Class for handling durations in all units (i.e. years, months, ...).
    In contrast, the datetime.timedelta only handles days, seconds and
    milliseconds. Which means the duration has always a definded extent in
    seconds.

    Here the arithmetic is completely different: A duration of a year can
    be 365 days, 366 days, even with 2 leap seconds included... Depends on the
    reference date the duration is applied.
    """

    def __init__(self):
        """
        Set up object object.
        """
        self.positive = True
        self.year = 0
        self.month = 0
        self.day = 0
        self.hour = 0
        self.minute = 0
        self.second = 0
        self.microsecond = 0

    def diff(self, time1, time2):
        """
        Calculates the ISO 8601 duration from the difference of two timestamps.
        Parameters:
            time1   timestam 1 (earlier)
            time2   timestamp 2 (later)
        Returns:
            None
        """
        if time2 < time1:
            self.positive = False
            hlp = time1
            time1 = time2
            time2 = hlp

        # Years
        offset = 0
        if time2.year > time1.year:
            for offset in (time2.year-time1.year, time2.year-time1.year-1):
                test = time1 + relativedelta(years=offset)
                # use relativedelta (replace year crashes on leap days!)
                if test <= time2:
                    time1 = test
                    break
        self.year = offset

        # Months
        offset = 0
        max_ = int(math.ceil((time2-time1).total_seconds()/60.0/60.0/24.0/28.0))
        min_ = math.trunc((time2-time1).total_seconds()/60.0/60.0/24.0/31.0)
        # min and max will always be different, except time1==time2
        for offset in reversed(list(range(min_, max_ + 1))):
            test = time1 + relativedelta(months=offset)
                # use relativedelta (replace year crashes on leap days!)
            if test <= time2:
                time1 = test
                break
        self.month = offset

        # Days
        offset = (time2-time1).days
        time1 = time1 + datetime.timedelta(days=offset)
        self.day = offset

        # Hours
        totalsec = (time2-time1).total_seconds()
        offset, totalsec = divmod(totalsec, 3600)
        self.hour = int(offset)
        # Minutes
        offset, totalsec = divmod(totalsec, 60)
        self.minute = int(offset)
        # Seconds
        offset, totalsec = divmod(totalsec, 1)
        self.second = int(offset)
        # Microseconds
        if totalsec != 0.0:
            self.microsecond = int(totalsec * 1000000 + 0.5)

    def format(self, fmt=1):
        """
        Return the formated output for the object.
        Parameters:
            fmt      format specification according to ISO8601:
                      1: standard format, e.g. P3Y6M4DT12H30M5S
                      2: alternative extended format, e.g. P0003-06-04T12:30:05
                      3: alternative shortened format, e.g. P00030604123005
        Returns:
            formatted string
        """
        if fmt == 1:
            #standard format:
            res = "P"
            if not self.positive:
                res += "-"
            if self.year:
                res += "{}Y".format(self.year)
            if self.month:
                res += "{}M".format(self.month)
            if self.day:
                res += "{}D".format(self.day)
            if self.hour or self.minute or self.second or self.microsecond:
                res += "T"
            if self.hour:
                res += "{}H".format(self.hour)
            if self.minute:
                res += "{}M".format(self.minute)
            if self.microsecond:
                res += "{}S".format(self.second+self.microsecond/1000000.0)
            elif self.second:
                res += "{}s".format(self.second)
            if res == "P":
                # nothing set yet
                res += "T0S"
            return res
        if fmt == 2:
            # alternative extended format (with -, T and :)
            res = "P{}{:04d}-{:02d}-{:02d}T{:02d}:{:02d}:{:02d}".format(
                "" if self.positive else "-", self.year, self.month, self.day,
                self.hour, self.minute, self.second)
            if self.microsecond:
                res += ".{:06d}".format(self.microsecond)
            return res
        if fmt == 3:
            # alternative shortened format (without -, T and :)
            res = "P{}{:04d}{:02d}{:02d}{:02d}{:02d}{:02d}".format(
                "" if self.positive else "-", self.year, self.month, self.day,
                self.hour, self.minute, self.second)
            if self.microsecond:
                res += ".{:6d}".format(self.microsecond)
            return res



"""
ebas/domain/base.py
$Id: base.py 2561 2020-12-07 23:09:30Z pe $

Base Classes for EBAS domain layer (entity objects)

History:
V.1.0.0  2012-10-14  pe  initial version

"""

from nilutility.datatypes import DataObject
from nilutility.datetime_helper import DatetimeInterval

# exceptions used in the domain module
class EbasDomainError(Exception):
    """
    Base class for exceptions.
    """
    pass


class EbasDomainObjCache(DataObject):
    """
    Object cache for domain objects (SU and DS).
    This mechanism is needed to avoid recursion in HDI objects.
    """
    def __init__(self):
        DataObject.__init__(self)

    def clear(self):
        """
        Clear all caches.
        Parameters:
            None
        Returns:
            None
        """
        for ele in list(self.keys()):
            self[ele] = []
            del self[ele]


class EbasDomainDbContext(DataObject):
    """
    Database Context for Ebas Domain Objects.
    """

    def __init__(self, dbh=None, time=DatetimeInterval(None, None), us_id=None,
                 export_filter=0,
                 state=None, diff=None, diff_ctx=None, cache=None):
        # pylint: disable-msg=R0913
        # R0913: Too many arguments
        """
        Initialize DB Context.
        Parameters:
            dbh     database handle (ebas.db layer)
            time    time interval for the context (only work on metadata and
                    date within this interval)
                    tuple (start, end), both datetime.datetime or None (+- inf)
            us_id   user id for access (access restrictions)
            export_filter
                    filter option for export of data from db to domain
                    bitfield:
                    EXPORT_FILTER_INCLUDE_900, EXPORT_FILTER_INCLUDE_INVALID
                    defined in ebas.doain.__init__
            state   the DB state for the context (only work on metadata and
                    data which are valid in this DB state)
                    datetime.datetime, None=current
            diff    only work on metadata and data that differ between state
                    and diff.
                    datetime.datetime, None=on all valid data
            diff_ctx
                    the EbasDomainDbContext object used for handling the object
                    layer for objects in "reversed" diff/state.
                    (should only be used by EbasDomainDbContext.__init__)
                    None=create new context if needed.
            cache   cache object to be used (None=create new cache)
        """
        DataObject.__init__(self)
        self.dbh = dbh
        self.time = time
        self.state = state
        self.diff = diff
        self.us_id = us_id
        self._dbwrite_obj = set()  # all objects which need to be written
        self._dbwrite_first = None  # object should be written first (useful?)
        self.export_filter = export_filter
        if cache:
            self.cache = cache
        elif dbh:
            self.cache = EbasDomainObjCache()
        if self.dbh:
            self.dbh.US_ID = us_id
        if self.diff:
            if not self.dbh:
                raise EbasDomainError("diff without dbh")
            if not self.state:
                self.state = self.dbh.safe_ts
            if diff_ctx:
                if diff_ctx.diff != self.state or \
                   diff_ctx.state != self.diff or diff_ctx.time != self.time:
                    raise EbasDomainError("diff_ctx: state/diff/time invalid")
                self.diff_ctx = diff_ctx
            else:
                # Create a "reversed" DbContext: stat/diff is reversed,
                # diff_ctx=self, and force a enew cache:
                self.diff_ctx = EbasDomainDbContext(
                    dbh=self.dbh, time=self.time, us_id=self.us_id,
                    export_filter=self.export_filter,
                    state=self.diff, diff=self.state, diff_ctx=self, cache=None)
        elif diff_ctx:
            raise EbasDomainError("diff_ctx without diff timestamp")

    def add_dbwrite_obj(self, obj):
        """
        Add object to the set of objects to be written on write_db.
        Parameters:
            obj    object to be added
        Returns:
            None
        """
        if not self._dbwrite_obj:
            self._dbwrite_first = obj
        if obj not in self._dbwrite_obj:
            self._dbwrite_obj.add(obj)

    def write_db(self):
        """
        Write all objects toi the DB.
        Parameters:
            None
        Returns:
            None
        """
        if self._dbwrite_first:
            self._dbwrite_first.write_db()
        for obj in self._dbwrite_obj:
            if obj != self._dbwrite_first:
                obj.write_db()
        self._dbwrite_obj = set()
        self._dbwrite_first = None



"""
ebas/nasa_ames/read
$Id: read.py 2097 2018-07-10 10:49:34Z pe $

File input functionality for EBAS NASA Ames module

! Attention !
Basic Nasa Ames 1001 functionality is implemented in the NasaAmes1001 class.
The ebas/nasa_ames (this) module is about EBAS extentions!

History:
V.1.0.0  2013-03-05  pe  initial version

"""

from .parse import NasaAmesPartialReadParser
from ..base import EbasNasaAmesReadError
from fileformats.NasaAmes1001 import NasaAmes1001, NasaAmes1001Error

class NasaAmesPartialRead(NasaAmesPartialReadParser):  # pylint: disable=R0901
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    """
    Nasa Ames I/O object.
    This is one part of the partial class NasaAmes (input functionality).
    """

    def _read(self, filename,  # pylint: disable=R0913, W0221
              ignore_rescode=False, ignore_revdate=False,  # @UnusedVariable
              ignore_parameter=False, ignore_sampleduration=False,
              skip_data=False,  # @UnusedVariable
              skip_unitconvert=False, skip_variables=None,  # @UnusedVariable
              encoding=None, ignore_numformat=False, ignore_dx0=False):
        # # @UnusedVariable
        # R0913 Too many arguments
        # W0221 _read:Arguments number differs from overridden '_read' method
        """
        Reads a NASA Ames files to NasaAmes object.
        Parameters:
            filename          path and filename
            ignore_rescode    ignore rescde errors (downgraded to warning)
            ignore revdate    ignore revdate errors (downgrade to warning)
            ignore_parameters ignore errors related to paramters and units
                              this is needed when a file with non standard
                              vnames should be processed without importing it
                              into the ebas.domain.
            skip_data         skip reading of data (speedup, if data are not
                              needed at application level). First and last line
                              are read. (Tjis is just passed to NasaAmes1001)
            skip_unitconvert  skips the automatic unit conversion on input
                              this is needed when a file with non standard units
                              should be processed without importing it into the
                              ebas.domain.
            skip_variables    list of variable numbers to be skipped (variabzle
                              numbers start with 1 for the first variable after
                              end_time)

            Specific for EbasNasaAmes._read:
            encoding          encoding of the file (passed through to
                              NasaAmes1001)
                              default: None (NasaAmes1001 tries to find a codec)
            ignore_numformat  ignore number format errors
            ignore_dx0        ignore error when DX=0 with regular data

        Returns:
            None
        Raises:
            IOError (from builtin, file open)
            EbasNasaAmesReadError
        """
        strictness = 0
        if not ignore_numformat:
            strictness |= NasaAmes1001.STRICT_NUMFORMAT
        if not ignore_dx0:
            strictness |= NasaAmes1001.STRICT_DX
        self.nasa1001 = NasaAmes1001(strictness=strictness)
        try:
            self.nasa1001.read(filename, encoding, skip_data=skip_data,
                               condense_messages=self.msg_condenser.threshold)
        except NasaAmes1001Error:
            pass
        self.errors += self.nasa1001.errors
        self.warnings += self.nasa1001.warnings
        if self.errors:
            self.logger.info("Exiting because of previous errors")
            raise EbasNasaAmesReadError(
                "{} Errors, {} Warnings".format(self.nasa1001.errors,
                                                self.nasa1001.warnings))

        self.parse_nasa_ames_1001(skip_variables=skip_variables,
                                  ignore_parameter=ignore_parameter)

    def msgref_vnum_lnum(self, vnum):
        """
        Get info on variable number and line number for error messages.
        Overrides the EbasFileBase method.
        If the file has been read from an input file, the the variable number
        and line numbers are adjusted to the actual file.
        Parameters:
            vnum    variable index in the file object
        Returns:
            (file_vnum, file_num) where:
                file_vnum: the real variable number in the physical file
                           (if there was a physical file read)
                           Variable 1 is the first variable after end_time
                file_lnum: the line number offset for the first data line in
                           the physical file (if there was a physical file read)
                file_metalnum: the line number reference that should be used for
                               metadata releted messages
        """
        file_vnum, file_lnum, file_metalnum = \
            super(NasaAmesPartialRead, self).msgref_vnum_lnum(vnum)
        # if read from file, get real variable number and line number in file:
        # (ibeacuse flag columns, possible skipped columns in between...)
        if self.internal.read_tmp and 'var_num' in self.internal.read_tmp \
           and self.internal.read_tmp.var_num:
            file_vnum = self.internal.read_tmp.var_num[vnum]
            file_metalnum = 13 + file_vnum
        if self.nasa1001:
            file_lnum = 13 + self.nasa1001.data.NV + 1 + \
                len(self.nasa1001.data.SCOML) + 1 + \
                len(self.nasa1001.data.NCOML)

        return (file_vnum, file_lnum, file_metalnum)



"""
ebas/nasa_ames/nasa_ames.py
$Id: nasa_ames.py 1527 2017-02-15 18:36:36Z pe $

EBAS NASA Ames module

! Attention !
Basic Nasa Ames 1001 functionality is implemented in the NasaAmes1001 class.
The ebas/nasa_ames (this) module is about EBAS extentions for Nasa Ames and
implements an interface to the ebas domain class (DB)

History:
V.1.0.0  2012-10-15  pe  initial version

"""

from .write import NasaAmesPartialWrite
from .read.read import NasaAmesPartialRead

class EbasNasaAmes(NasaAmesPartialWrite,   #pylint: disable-msg=R0901,R0904
                   NasaAmesPartialRead):
    # R0901: Too many ancestors
    # R0904: Too many public methods
    # This is a quite big partial class, therefore we ignore this warning
    """
    Nasa Ames I/O object.
    This is a partial class distributed over many source files.
    """




"""
$Id: haversine.py 1809 2017-10-17 10:22:58Z pe $

Geospatianls functions - Haversine formula
Spherical trigonometry

History:
V.1.0.0  2017-10-16  pe  initial version
"""

import math

UNIT_FACTORS = {
    "nm": 1,
    "km": 1.852,
    "m": 1852,
}

def haversine_dist(pos1, pos2, unit='nm'):
    """
    Haversine great circle distance between two points.
    Parameters:
        pos1, pos2   the two position, each (lat, lon)
        unit         result unit ('km', 'nm')
    Returns:
        distance
    Raises:
        ValueError  on unsupported unit
    """
    r_nm = 180*60/math.pi
    try:
        rad = r_nm * UNIT_FACTORS[unit]
    except KeyError:
        raise ValueError("unsupported unit '{}'".format(unit))
    lat1, lon1, lat2, lon2 = [math.radians(x) for x in list(pos1) + list(pos2)]
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a__ = math.sin(dlat/2)**2 + math.cos(lat1) * math.cos(lat2) * \
        math.sin(dlon/2)**2
    c__ = 2 * math.asin(math.sqrt(a__))
    return c__ * rad



"""
ebas/io/xmlwrap.py
$Id: xmlwrap.py 1527 2017-02-15 18:36:36Z pe $

EBAS I/O module, xmlwrap
"""

import sys

def xmlwrap_header():
    """
    Writes the xmlwrap-header.
    """
    sys.stdout.write('<?xml version="1.0" encoding="UTF-8" '
                    'standalone="yes"?>\n')
    sys.stdout.write('<Fileset>\n')

def xmlwrap_trailer():
    """
    Writes the xmlwrap-trailer.
    """
    sys.stdout.write('</Fileset>\n')

def xmlwrap_fileheader(filename):
    """
    Writes the xmlwrap-fileheader.
    """
    sys.stdout.write(' <File filename="{}">\n'.format(filename))
    sys.stdout.write('  <![CDATA[')
    # no newline

def xmlwrap_filetrailer():
    """
    Writes the xmlwrap-filetrailer.
    """
    sys.stdout.write(']]>\n')
    sys.stdout.write(' </File>\n')



"""
NOAA NMHC Flask data file format

$Id: __init__.py 1796 2017-10-11 11:53:47Z pe $
"""

from .base import NOAA_NMHC_Flask_Error, NOAA_NMHC_Flask_Inconsistent
from .read import NOAA_NMHC_Flask_ReadError

from .read import NOAA_NMHC_Flask_Read, NOAA_NMHC_Flask_Rawdata_Read, \
    NOAA_NMHC_Flask_Header_Read, NOAA_NMHC_Flask_Data_Read, \
    NOAA_NMHC_Flask_HeaderContact_Read, NOAA_NMHC_Flask_HeaderDescription_Read

class NOAA_NMHC_Flask_Rawdata(NOAA_NMHC_Flask_Rawdata_Read):
    """
    Construct the final class.
    """
    pass

class NOAA_NMHC_Flask_HeaderContact(NOAA_NMHC_Flask_HeaderContact_Read):
    """
    Construct the final class.
    """
    pass

class NOAA_NMHC_Flask_HeaderDescription(NOAA_NMHC_Flask_HeaderDescription_Read):
    """
    Construct the final class.
    """
    pass

class NOAA_NMHC_Flask_Header(NOAA_NMHC_Flask_Header_Read):
    """
    Construct the final class.
    """
    pass

class NOAA_NMHC_Flask_Data(NOAA_NMHC_Flask_Data_Read):
    """
    Construct the final class.
    """
    pass


class NOAA_NMHC_Flask(NOAA_NMHC_Flask_Read):
    """
    Construct the final class.
    """
    pass



"""
$Id: bc.py 2597 2021-01-19 13:36:24Z pe $

EBAS Masterdata Class for boundary check masterdata (BC and BA)

This module implements the class EbasMasterBC.

History:
V.1.0.0  2016-10-12  pe  initial version

"""

from .offline_masterdata import BCOfflineMasterData
from .base import EbasMasterBase

class EbasMasterBCNoProj(Exception):
    """
    Exception class, raise in case of no project specified in BC selection.
    """
    pass

class EbasMasterBCDuplicateError(Exception):
    """
    Exception class, raise in case of duplicate rules in BC
    """
    pass

class EbasMasterBCDuplicateErrorList(Exception):
    """
    Exception class, raise in case of duplicate rules in BC.
    Error list, generated in select() (possibly one error per project).
    """
    def __init__(self, message, errlist, retlist):  # pylint: disable=W0235
        # W0235: Useless super delegation in method
        # --> just for documenting what argument 0, 1, and 2 are used for
        """
        Initialize instance.
        Parameters:
            message    general exception text
            errlist    list of exceptions raised by _select_by_proj()
            retlist    list of bc rules found for projects w/o error
        """
        super(EbasMasterBCDuplicateErrorList, self).__init__(message, errlist,
                                                             retlist)

    @property
    def errlist(self):
        """
        Property method for errlist.
        Parameters:
            None
        Returns:
            errlist
        """
        return self.args[1]

    @property
    def retlist(self):
        """
        Property method for retlist.
        Parameters:
            None
        Returns:
            retlist
        """
        return self.args[2]


class EbasMasterBC(EbasMasterBase, BCOfflineMasterData):
    """
    Domain Class for boundary check masterdata (BA and BA)
    Objects of this class do not represent entities, this class provides class
    methods for handling boundary check rules and selecting applicable boundary
    checks.
    Boundary check master data are retrieved from database or from offline
    storage when no database access is possible.
    """

    # read offline masterdata for BC and BA (boundary checks)
    # Those are fallback values, will be read from database as soon as possible.
    BCOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read ac masterdata if dbh is provided.
        """
        BCOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)

    def select(self, crit_dict):
        """
        Select the best matching boundary check for a set of dataset metadata.
        Parameters:
            crit_dict    criteria dict (metadata for the dataset). Needed keys:
                         RE_REGIME_CODE, MA_MATRIX_NAME, CO_COMP_NAME,
                         PR_ACRONYM [list],
                         ST_STATION_CODE, FT_TYPE, FI_REF, DL_DATA_LEVEL,
                         SC_STATISTIC_CODE and characteristics
        Returns:
            List of boundary check rules (max one for each project), free of
            duplicates. Empty if no rules apply.
        Raises:
            EbasMasterBCDuplicateError in case of multiple rules matching per
            project
        """
        # Select all possible candidates
        # - regime, matrix, component MUST match exactly (they are not NULL in
        # the DB because the triple defines the parameter and thus the unit,
        # otherwise a boundery without defined unit does not make sense)
        # - most other criteria must either match or be set to None in BC
        # - special case characteristics: they must either be an empty list in
        # BC (no characteristics defined), or the complete set of
        # characteristics defined in BC must match.
        #
        # PR_ACRONYM overrule all others for sure. If a project defines
        # boundaries for a parameter it should be chosen over an exceptional
        # rule for station, instrument etc without project specified.
        # Exceptions for a project need to be defined explicidly.(with
        # project set)
        # Another complication is that a time series may be associated to
        # several projects. For each of those projects, the best fitting rule
        # is selected. In the end, for each check limit, the most restrictive
        # is selected to create the overal limits to be used.
        #
        # The rest of the priority chain is in reverse order of how restrictive
        # the criteria usually will be: FI_REF is very specific compared to
        # ST_STATION_CODE, and so on.
        #
        # for each project, find the best fitting rule:
        ret = []
        raise_ = []
        if 'PR_ACRONYM' not in crit_dict or not crit_dict['PR_ACRONYM']:
            raise EbasMasterBCNoProj('No framework specified')
        for proj in crit_dict['PR_ACRONYM']:
            try:
                proj_rule = self._select_by_proj(crit_dict, proj)
            except EbasMasterBCDuplicateError as excpt:
                raise_.append(excpt)
            else:
                if proj_rule and proj_rule not in ret:
                    ret.append(proj_rule)
        if raise_:
            raise EbasMasterBCDuplicateErrorList(
                'duplicate boundary check rules for dataset ({})'.format(
                    crit_dict), raise_, ret)
        return ret

    def _select_by_proj(self, crit_dict, pr_acronym):
        """
        Select the best matching boundary check for a set of dataset metadata
        and A SPECIFIC PROJECT ACRONYM.
        Parameters:
            crit_dict    criteria dict (metadata for the dataset). Needed keys:
                         RE_REGIME_CODE, MA_MATRIX_NAME, CO_COMP_NAME,
                         ST_STATION_CODE, FT_TYPE, FI_REF, DL_DATA_LEVEL,
                         SC_STATISTIC_CODE and characteristics
                         (PR_ACRONYM in the crit_dict is ignored, instead the
                         pr_acronym parameter is used (only one project))
            pr_acronym   project acronym for which the best fitting rule should
                         be found
        Returns:
            best matching BC dictionary for this project and the criteria
            metadata. If there is no project specific rule, the best matching
            general rule is returned.
            Thus, a project rule can both increase or decrease the general
            quality requirements. Exceptions (e.g. for a single station must
            then be done for all defined projects again.
            None in case of no match
        Raises:
            EbasMasterBCDuplicateError in case of multiple rules matching per
            project
        """
        cand = [bc_ for bc_ in sorted(list(self.META.values()),
                                      key=lambda x: x['BC_ID'])
                if bc_['RE_REGIME_CODE'] == crit_dict['RE_REGIME_CODE'] and \
                    bc_['MA_MATRIX_NAME'] == crit_dict['MA_MATRIX_NAME'] and \
                    bc_['CO_COMP_NAME'] == crit_dict['CO_COMP_NAME'] and \
                    bc_['PR_ACRONYM'] in (None, pr_acronym) and \
                    bc_['ST_STATION_CODE'] in \
                        (None, crit_dict['ST_STATION_CODE']) and \
                    bc_['FT_TYPE'] in (None, crit_dict['FT_TYPE']) and \
                    bc_['FI_REF'] in (None, crit_dict['FI_REF']) and \
                    bc_['DL_DATA_LEVEL'] in \
                        (None, crit_dict['DL_DATA_LEVEL']) and \
                    bc_['SC_STATISTIC_CODE'] in \
                        (None, crit_dict['SC_STATISTIC_CODE']) and \
                    (not bc_['characteristics'] or \
                     all([ct_ in crit_dict['characteristics'] and \
                          bc_['characteristics'][ct_] == \
                              crit_dict['characteristics'][ct_]
                          for ct_ in bc_['characteristics']]))]

        for prio in ('PR_ACRONYM', 'FI_REF', 'ST_STATION_CODE', 'FT_TYPE',
                     'DL_DATA_LEVEL', 'SC_STATISTIC_CODE'):
            if len(cand) == 1:
                return cand[0]
            if prio == 'PR_ACRONYM':
                newcand = [bc_ for bc_ in cand
                           if bc_['PR_ACRONYM'] == pr_acronym]
            else:
                newcand = [bc_ for bc_ in cand if bc_[prio] == crit_dict[prio]]
            if newcand:
                cand = newcand
        if len(cand) == 1:
            return cand[0]
        if cand:
            # still more then one candidate: count number of characteristics
            # matches for each candidate:
            cnt = [[ct_ in crit_dict['characteristics'] and
                        bc_['characteristics'][ct_] == \
                            crit_dict['characteristics'][ct_]
                    for ct_ in bc_['characteristics']].count(True)
                   for bc_ in cand]
            cand = [cand[i] for i in range(len(cnt)) if cnt[i] == max(cnt)]
        if len(cand) == 1:
            return cand[0]
        if cand:
            # still more then one candidate? Looks like we have identical rules?
            raise EbasMasterBCDuplicateError(
                'duplicate boundary check rules for dataset ({}), framework {}'
                .format(crit_dict, pr_acronym))
        return None



"""
Module generateDS_ebas: With generateDS.py generated python data structures (for example, class definitions)
                        from ebas.xsd XML Schema document.
                        generatedssuper.py is imported by generateDS_ebas and contains ebas specific implementations.
"""

# import ebas
# import gml


"""
Ebas IO base classes

$Id: base.py 1528 2017-02-15 18:45:40Z pe $
"""

class EbasIOError(Exception):
    """Error base class for EBAS IO."""
    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.msg = msg
    def __str__(self):
        return self.msg


"""
Module atmos_phys.convert.aerosol: Unit conversion routines for aerosols.
"""

from nilutility import numeric
from nilutility.datatypes import DataObject
from .base import ConvertBaseFactor
from .elements import ELEMENTS

class ConvertNitrate(ConvertBaseFactor):
    """
    Conversion class for nitrate (NO3) conversions:
    Mass concentrationis "ug/m3" <-> "ug N/m3".

    Attributes:
        from_unit: The source unit of conversion
        to_unit: The result unit after conversion
        parameters: Parameters for the conversion
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        if self.from_unit not in ('ug/m3', 'ug N/m3') or \
           self.to_unit not in ('ug/m3', 'ug N/m3') or \
           self.from_unit == self.to_unit:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

        self.parameters = DataObject(factor=None)
        if self.to_unit == 'ug N/m3':
            self.parameters.factor = \
                ELEMENTS['N']['MolarMass'] / \
                (ELEMENTS['N']['MolarMass'] + 3*ELEMENTS['O']['MolarMass'])
        else:
            self.parameters.factor = \
                (ELEMENTS['N']['MolarMass'] + 3*ELEMENTS['O']['MolarMass']) / \
                ELEMENTS['N']['MolarMass']


class ConvertNitrite(ConvertBaseFactor):
    """
    Conversion class for nitrite (NO2) conversions:
    Mass concentrationis "ug/m3" <-> "ug N/m3".
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        if self.from_unit not in ('ug/m3', 'ug N/m3') or \
           self.to_unit not in ('ug/m3', 'ug N/m3') or \
           self.from_unit == self.to_unit:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

        self.parameters = DataObject(factor=None)
        if self.to_unit == 'ug N/m3':
            self.parameters.factor = \
                ELEMENTS['N']['MolarMass'] / \
                (ELEMENTS['N']['MolarMass'] + 2*ELEMENTS['O']['MolarMass'])
        else:
            self.parameters.factor = \
                (ELEMENTS['N']['MolarMass'] + 2*ELEMENTS['O']['MolarMass']) / \
                ELEMENTS['N']['MolarMass']


class ConvertAmmonium(ConvertBaseFactor):
    """
    Conversion class for ammonium (NH4) conversions:
    Mass concentrationis "ug/m3" <-> "ug N/m3".
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        if self.from_unit not in ('ug/m3', 'ug N/m3') or \
           self.to_unit not in ('ug/m3', 'ug N/m3') or \
           self.from_unit == self.to_unit:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

        self.parameters = DataObject(factor=None)
        if self.to_unit == 'ug N/m3':
            self.parameters.factor = \
                ELEMENTS['N']['MolarMass'] / \
                (ELEMENTS['N']['MolarMass'] + 4*ELEMENTS['H']['MolarMass'])
        else:
            self.parameters.factor = \
                (ELEMENTS['N']['MolarMass'] + 4*ELEMENTS['H']['MolarMass']) / \
                ELEMENTS['N']['MolarMass']


class ConvertSulphate(ConvertBaseFactor):
    """
    Conversion class for nitrate (NO3) conversions:
    Mass concentrationis "ug/m3" <-> "ug N/m3".
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        if self.from_unit not in ('ug/m3', 'ug S/m3') or \
           self.to_unit not in ('ug/m3', 'ug S/m3') or \
           self.from_unit == self.to_unit:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

        self.parameters = DataObject(factor=None)
        if self.to_unit == 'ug S/m3':
            self.parameters.factor = \
                ELEMENTS['S']['MolarMass'] / \
                (ELEMENTS['S']['MolarMass'] + 4*ELEMENTS['O']['MolarMass'])
        else:
            self.parameters.factor = \
                (ELEMENTS['S']['MolarMass'] + 4*ELEMENTS['O']['MolarMass']) / \
                ELEMENTS['S']['MolarMass']




"""
$Id: opendap.py 1625 2017-05-23 11:25:04Z pe $

EBAS OPeNDAP module
"""

from .write import EbasOPeNDAPPartialWrite

class EbasOPeNDAP(EbasOPeNDAPPartialWrite):
    """
    EBAS I/O OPeNDAP.
    This is a partial class (only write is implemented for OPeNDAP).
    """
    pass
